__metaclass__ = type

from contextlib import contextmanager
from datetime import datetime
import errno
import fcntl
from jenkins import JenkinsException
import json
import logging
import os
import subprocess

import yaml

from jobs import (
    BUILD_REVISION,
    PUBLISH_REVISION,
)
from utility import (
    archive_version_url,
    run_s3cmd,
    temp_dir,
)


BUILDING = 'building'
SUCCEEDED = 'succeeded'
FAILED = 'failed'
PENDING = 'pending'
TERMINAL_STATUS = (SUCCEEDED, FAILED)


BLESS = 'bless'
CURSE = 'curse'
FINAL_RESULTS = (BLESS, CURSE)


def open_rw(filename):
    """Open a file for reading and writing.

    The file is created if it does not exist (like w+), but is not truncated
    (like r+).
    """
    fd = os.open(filename, os.O_RDWR | os.O_CREAT)
    return os.fdopen(fd, 'w+')


class StateFileInUse(Exception):
    """Raised when the lock for the state file is held by another process."""
    pass


class StateFile:
    """Provide programmatic access to the state data."""

    def __init__(self, data=None, fileobj=None):
        if data is None:
            data = {
                'versions': {},
                'versions_by_revision': {},
                'test_builds': {},
            }
        self._data = data
        self._fileobj = fileobj
        self.logger = logging.getLogger('cidirector')

    @classmethod
    def from_file(cls, fileobj):
        """Return a StateFile for a specified File object."""
        data = yaml.safe_load(fileobj)
        fileobj.seek(0)
        return cls(data, fileobj)

    @classmethod
    @contextmanager
    def acquire(cls):
        """Context manager that provides a StateFile and file _object.

        An exclusive lock is taken on the file.  If the lock cannot be
        acquired, StateFileInUse is raised.
        """
        state_file_obj = open_rw(StateFile.file_name())
        with state_file_obj:
            try:
                fcntl.lockf(state_file_obj, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except IOError as e:
                if e.errno != errno.EAGAIN:
                    raise
                raise StateFileInUse('State file in use.')
            try:
                yield cls.from_file(state_file_obj), state_file_obj
            finally:
                state_file_obj.flush()
                fcntl.lockf(state_file_obj, fcntl.LOCK_UN)

    @staticmethod
    def file_name():
        """Return the filename of the current user's state file."""
        return os.path.join(os.environ['HOME'], '.config/ci-director-state')

    @property
    def current_version_number(self):
        """Return the build number of the current version."""
        return self._data.get('current_version')

    @property
    def _current_version(self):
        current_version = self._data.get('current_version')
        if current_version is None:
            return None
        return self._data['versions'][current_version]

    def start_revision(self, branch, revno, revision_id, release_number,
                       build_id):
        """Record that building a revision has started.

        :param branch: Location of the branch being built.
        :param revno: Revno of the branch being built.
        :param revision_id: Revision-id of the branch being built.
        :param release_number: Version number of the Juju release.
        :param build_id: Number of the build.
        """
        assert build_id not in self._data['versions']
        self._data['current_version'] = build_id
        self._data['versions'][build_id] = {
            'branch': branch,
            'revno': revno,
            'revision_id': revision_id,
            'release_number': release_number,
            'tests': {},
            'build': {},
            'started': datetime.utcnow().isoformat() + 'Z',
        }
        builds = self._data['versions_by_revision'].setdefault(revision_id, [])
        builds.append(build_id)
        self.get_build_revision_job().start_build(build_id)

    def finish_revision(self, now):
        info = self.get_current_version()
        info['finished'] = now.isoformat() + 'Z'

    def get_current_version_summary(self):
        """Retrieve a summary of the current revision.

        :return: a dict of build_number, revno, branch, revision_id.  (Same
            meaning as in start_revision).
        """
        current_version = self._current_version
        result = dict((k, current_version[k]) for k in
                      ['revno', 'branch', 'revision_id', 'release_number'])
        result['build_number'] = self._data['current_version']
        revno = current_version.get('revno')
        if revno is None:
            result['short_revision'] = current_version['revision_id'][:8]
        else:
            result['short_revision'] = 'r%d' % revno
        return result

    def get_build_revision_job(self):
        """Return the StateFileJob for build-revision."""
        version_data = self._current_version
        if version_data is None:
            return None
        # Handle legacy entries-- new entries will always have this set by
        # start_revision()
        build_data = version_data.setdefault(
            'build', {'status': PENDING, 'building': True})
        return StateFileJob(BUILD_REVISION, build_data, logger=self.logger)

    def publication_job(self):
        """Return the publication StateFileJob for the current version.

        Return None if there isn't a current version.
        """
        version_data = self._current_version
        if version_data is None:
            return None
        for job in self.iter_grouped_jobs():
            if job.test_id == PUBLISH_REVISION:
                return job
        tags = (
            'subjects=release,feature function=publish-metadata '
            'arch=all series=all substrate=none')
        return self.get_grouped_job(
            PUBLISH_REVISION, group_name='release-tests', failure_threshold=2,
            tags=tags)

    def get_grouped_job(self, job_id, group_name=None, failure_threshold=None,
                        tags=None, vote=True):
        """Retrieve the StateFileJob for a given grouped job."""
        if self._current_version is None:
            return None
        if group_name is None:
            group = self._current_version
        else:
            group = self._current_version.setdefault(group_name, {})
        default = {
            'status': PENDING,
            'tags': tags,
            'failures': 0,
            'vote': vote,
        }
        job_dict = group.setdefault(job_id, default)
        return StateFileJob(job_id, job_dict, failure_threshold, vote=vote,
                            logger=self.logger)

    def iter_grouped_jobs(self):
        """Iterate through all group StateFileJobs for the current version.

        WARNING: This method yields the discovered job, some jobs maybe
        created in the future.
        """
        if 'current_version' not in self._data:
            return
        for job_id, job_dict in self._current_version.items():
            if job_id == 'build' or not isinstance(job_dict, dict):
                continue
            if 'failures' in job_dict:
                yield StateFileJob(
                    job_id, job_dict, None,
                    vote=job_dict.get('vote', True), logger=self.logger)
            else:
                group = job_dict
                for job_id, job_dict in group.items():
                    if isinstance(job_dict, dict) and 'failures' in job_dict:
                        yield StateFileJob(
                            job_id, job_dict, None,
                            vote=job_dict.get('vote', True),
                            logger=self.logger)

    def revision_prepared(self, revision_id):
        """Determine whether a revision has already been prepared.

        :param revision_id: The revision to consider
        :return: True or False
        """
        builds = self._data['versions_by_revision'].get(revision_id, [])
        return bool(len(builds) > 0)

    def write(self):
        """Write this state data to the supplied file."""
        self._fileobj.seek(0)
        yaml.safe_dump(self._data, self._fileobj)
        self._fileobj.truncate()

    def save_result(self, status, build):
        """"Record the bless/curse result for the current version.

        :param status: BLESS or CURSE
        :param build: The number of the build where this was recorded.
        """
        if status not in FINAL_RESULTS:
            raise ValueError('Illegal value: %s' % status)
        self._current_version['result'] = {'status': status, 'build': build}

    def get_result(self):
        """Retrieve the bless/curse result for the current version.

        :return: a tuple of (status, build), or (None, None) if no result was
            recorded.
        """
        result = self._current_version.get('result')
        if result is None:
            return None, None
        return result['status'], result['build']

    def get_complete_failure_info(self, revision_build):
        """Return the complete information about failures for a revision build.

        Only those failures that contribute to the build being cursed are
        reported, not -devel failures.
        """
        revision_data = self._data['versions'][revision_build]
        return revision_data.get('complete_failure_info')

    def set_complete_failure_info(self, revision_build, failure_info):
        """Store the complete information about failures for a revision build.

        Only those failures that contribute to the build being cursed are
        stored, not -devel failures.
        """
        revision_data = self._data['versions'][revision_build]
        revision_data['complete_failure_info'] = failure_info

    def get_current_version(self):
        return self._current_version

    def build_info_to_s3(self):
        dest_url = '%s/result.json' % archive_version_url(
            self.current_version_number)
        with temp_dir() as work_dir:
            file_path = os.path.join(work_dir, 'result.json')
            with open(file_path, 'w') as result:
                info = self.get_current_version()
                json.dump(info, result)
            run_s3cmd(['--no-progress', 'put', file_path, dest_url])


class StateFileJob:
    """Representation of a Job in the StateFile."""

    def __init__(self, test_id, data, failure_threshold=1, vote=True,
                 logger=None):
        """Constructor.

        :param test_id: The ID of the job.
        :param data: The StateFile dict representing the job.
        :param failure_threshold: The number of build failures that cause the
            job to be considered a failure.
        """
        self.test_id = test_id
        # The failure_threshold comes from the data, but when absent, the
        # value passed to init is used, or finally fall back to 1.
        threshold = failure_threshold or 1
        # SFJ is a light-wrapper around the data dict. _data must be a
        # reference to the data dict because it is just a dict in a larger
        # dict. This data dance ensures the data has consistent keys with
        # sane defaults.
        base_data = {
            'failure-threshold': threshold,
            'failures': 0,
            'last_build': None,
            'status': PENDING,
            'vote': vote,
            'building': False,
        }
        base_data.update(data)
        data.update(base_data)
        self._data = data
        self.logger = logger
        self.vote = vote

    def __repr__(self):
        return 'StateFileJob<%r, %r>' % (self.test_id, self.get_status())

    def _summary(self):
        return (self.test_id,
                tuple(sorted((k, v) for k, v in self._data.items())),
                self.failure_threshold)

    def __hash__(self):
        return hash(self._summary())

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        return self._summary() == other._summary()

    @property
    def build_number(self):
        return self._data['last_build']

    @property
    def failure_threshold(self):
        return self._data['failure-threshold']

    def start_build(self, build_id):
        """Record a build of a provider test.

        The initial status is BUILDING.
        :param build_id: The build number of the build.
        """
        self._data.update({
            'last_build': build_id,
            'building': True,
        })

    def get_status(self):
        """Return the current status of a job.

        If the current build is still running, return BUILDING.

        If the failure count reaches the threshold, and the status is
        PENDING, return FAILED.  Otherwise, return the 'status' value.
        """
        if self._data['building']:
            return BUILDING
        status = self._data['status']
        if (status == PENDING
                and self._data['failures'] >= self.failure_threshold):
            return FAILED
        return status

    def update_from_jenkins(self, jenkins, current_version):
        """Update to the last build's result via Jenkins.

        :param jenkins: A jenkins.Jenkins instance
        """
        last_build = self._data['last_build']
        if last_build is None:
            return
        try:
            info = jenkins.get_build_info(self.test_id, last_build)
        except JenkinsException as err:
            if str(err).endswith('number[%s] does not exist' % last_build):
                # A build can really disappear, see bug 1365808.
                self._data['last_build'] = None
                self._data['building'] = False
                if self.logger is not None:
                    self.logger.warn(
                        'Build %s of job %s is missing' % (
                            last_build, self.test_id))
                return
            else:
                raise
        result = info.get('result')
        if result is None:
            return
        old_status = self.get_status()
        self.update_from_build_result(result)
        if old_status == BUILDING and old_status != self.get_status():
            self.archive_results(
                info['artifacts'], current_version, info['url'], last_build)

    def update_from_build_result(self, result):
        """Update a provider test from its build result.

        Only FAILURE causes the failure count to increment.
        SUCCESS changes the status to SUCCEEDED, and other results change the
        status to PENDING.
        """
        def handle_FAILURE():
            self._data.update({
                'failures': self._data['failures'] + 1,
            })

        def handle_ABORTED():
            pass

        def handle_SUCCESS():
            self._data['status'] = SUCCEEDED

        handle_UNSTABLE = handle_FAILURE
        handle_NOT_BUILT = handle_ABORTED
        try:
            handler = locals()['handle_' + result]
        except KeyError:
            raise ValueError('No such status: %s' % result)
        self._data['building'] = False
        return handler()

    def archive_results(self, artifacts, current_version, build_url,
                        build_number):
        """Copy artifacts of this job to an S3 bucket."""
        if len(artifacts) == 0:
            return
        with temp_dir() as work_dir:
            dest_url = '%s/%s/build-%s/' % (
                archive_version_url(current_version), self.test_id,
                build_number)
            for artifact in artifacts:
                filename = artifact['fileName']
                if filename == 'empty':
                    continue
                relative_path = artifact['relativePath']
                source_url = '%s/artifact/%s' % (
                    build_url, relative_path)
                local_path = os.path.join(work_dir, filename)
                subprocess.check_call([
                    'wget', '-q', '-O', local_path, source_url])
            run_s3cmd(['--no-progress', 'sync', '%s/' % work_dir, dest_url])
