__metaclass__ = type

from contextlib import contextmanager
import logging
import os
import sys
from textwrap import dedent
from unittest import TestCase

from bzrlib.bzrdir import BzrDir
from jenkins import (
    auth_headers,
    Jenkins,
    JenkinsException,
)
from mock import patch

from cidirector.cidirector import (
    BuildRevisionJob,
    CIDirector,
    ConfigReader,
    get_build_parameters,
    get_buildvars_file,
    Job,
    JobInfo,
    main,
    Mailer,
    parse_buildvars_file,
    RecordResultJob,
    ResourcefulJob,
    ResultJudge,
    ServerInfo,
)
from cidirector.jobs import (
    BUILD_REVISION,
    PUBLISH_REVISION,
    JENKINS_BUILD_RESULTS,
    REVISION_RESULTS,
)
from cidirector.storage import (
    BLESS,
    BUILDING,
    CURSE,
    FAILED,
    PENDING,
    SUCCEEDED,
    StateFile,
    StateFileJob,
)
from cidirector.tests.utility import (
    finally_fail_job,
)
from cidirector.utility import (
    temp_dir,
)


WORKSPACE = 'workspace'


@contextmanager
def bzr_sandbox():
    with temp_dir() as new_home:
        with patch.dict(os.environ, {'BZR_HOME': new_home}):
            yield new_home


def make_bzr_tree(bzr_home, name, version, commits=1):
    path = os.path.join(bzr_home, name)
    tree = BzrDir.create_standalone_workingtree(path)
    version_dir = os.path.join(path, 'version')
    os.mkdir(version_dir)
    with open(os.path.join(version_dir, 'version.go'), 'w') as version_file:
        version_file.write('foo\nconst version = "%s" foobar\n' % version)
    tree.add(['version', 'version/version.go'])
    for unused in range(commits):
        revision_id = tree.commit('message', committer='foo')
    return tree, path, revision_id


def make_server_info(phase, buildable=None, building=None, disabled=None):
    if building is None:
        building = []
    if disabled is None:
        disabled = []
    jobs = []
    if phase == 'idle':
        pass
    if phase == 'preparing':
        building.append(BUILD_REVISION)
    if phase == 'testing':
        building.append(PUBLISH_REVISION)
    if buildable is not None:
        for job in buildable:
            jobs.append({'name': job, 'color': 'red'})
    for job in building:
        jobs.append({
            'name': job,
            'color': 'red_anime',
        })
    for job in disabled:
        jobs.append({
            'name': job,
            'color': 'disabled',
        })
    return ServerInfo({'jobs': jobs})


def log_calls(func):
    def wrapper(self, *args, **kwargs):
        self.calls.setdefault(func.__name__, []).append((args, kwargs))
        return func(self, *args, **kwargs)
    return wrapper


class FakeJenkins:

    server = 'fakettp://fake.fake/'

    def __init__(self, jobs, current_version=None):
        self.calls = {}
        self.job_info = {}
        self.buildvars = {}
        self.info = {'jobs': []}
        if current_version is not None:
            jobs[BUILD_REVISION] = {'build_number': current_version}
        self.build_info = {}
        self.instant_build = True
        for key, value in jobs.items():
            if value['build_number'] is None:
                last_build = None
            else:
                last_build = {'number': value['build_number']}
            self.job_info[key] = {
                'lastBuild': last_build,
                'description': value.get('description'),
            }
            building = value.get('building', False)
            color_suffix = '_anime' if building else ''
            self.info['jobs'].append({
                'name': key,
                'color': value.get('color', 'purple') + color_suffix,
            })
            test_dict = self.build_info.setdefault(key, {})
            test_dict[value['build_number']] = {
                'result': None if building else 'SUCCESS',
                'building': building,
                'number': value['build_number'],
                'artifacts': [],
                'url': self._make_build_url(key, value['build_number']),
            }
        self.queue = []

    def _make_build_url(self, job_name, build_number):
        return '%sjob/%s/%s' % (self.server, job_name, build_number)

    @log_calls
    def get_queue_info(self):
        return self.queue

    def enqueue(self, jobs):
        self.queue.extend({'task': {'name': job}} for job in jobs)

    @log_calls
    def build_job(self, name, parameters=None):
        info = self._get_job_info(name)
        if info['lastBuild'] is None:
            info['lastBuild'] = {'number': 1}
        else:
            info['lastBuild']['number'] = info['lastBuild']['number'] + 1
        build_number = info['lastBuild']['number']
        self.build_info[name][build_number] = {
            'result': None,
            'building': True,
            'number': build_number,
            'artifacts': [],
            'url': self._make_build_url(name, build_number),
        }

    @log_calls
    def get_build_info(self, test_id, build_id):
        if build_id in self.build_info[test_id]:
            return self.build_info[test_id][build_id]
        else:
            raise JenkinsException(
                'job[%s] number[%s] does not exist' % (test_id, build_id))

    def set_build_result(self, job_id, build_number, result, artifacts=[]):
        info = self.build_info.setdefault(job_id, {build_number: {}})
        build_info = info.setdefault(build_number, {})
        build_info['result'] = result
        build_info['url'] = self._make_build_url(job_id, build_number)
        build_info['artifacts'] = artifacts

    @log_calls
    def get_info(self):
        return self.info

    @log_calls
    def get_job_info(self, job_name):
        return self._get_job_info(job_name)

    def _get_job_info(self, name):
        try:
            return self.job_info[name]
        except KeyError:
            raise JenkinsException('job[%s] does not exist' % name)

    def jenkins_open(self, req):
        if (req.get_full_url() ==
                'fakettp://fake.fake/job/build-revision/ws/buildvars.bash'):
            return self.buildvars[WORKSPACE]


class TestGetBuildParameters(TestCase):

    def test_get_build_parameters(self):
        self.assertEqual(get_build_parameters({}), {})
        self.assertEqual(get_build_parameters({'actions': []}), {})
        self.assertEqual(get_build_parameters(
            {'actions': [{'parameters': []}]}), {})
        self.assertEqual(get_build_parameters(
            {'actions': [{'parameters': [
                {'name': 'foo', 'value': 6}]}]}), {'foo': 6})
        self.assertEqual(get_build_parameters({'actions': [
            {'parameters': [{'name': 'foo', 'value': 6}]},
            {'parameters': [
                {'name': 'bar', 'value': 5},
                {'name': 'baz', 'value': 4},
            ]},
        ]}), {'foo': 6, 'bar': 5, 'baz': 4})


class TestCIDirector(TestCase):

    def test_get_active_jobs(self):
        si = make_server_info('idle', building=['foo', 'bar'])
        director = CIDirector(FakeJenkins({}), StateFile())
        director.jenkins.enqueue(['baz', 'qux'])
        self.assertEqual(['foo', 'bar', 'baz', 'qux'],
                         director.get_active_jobs(si))

    def make_schedule_builds_director(self, branches, active=True,
                                      published=True, server_info=None):
        state_file = StateFile()
        state_file.start_revision('a', 1, 'b', '1.6', 2)
        if published:
            state_file.publication_job().update_from_build_result('SUCCESS')
        version = 2 if active else 3
        jenkins_dict = {
            PUBLISH_REVISION: {
                'build_number': 5,
                'description': '[ci-director]\n requires: build-revision'},
            REVISION_RESULTS: {'build_number': 1},
            'aws-deploy': {
                'build_number': 5,
                'description': '[ci-director]\n requires: build-revision'},
        }
        if server_info is not None:
            for entry in server_info.info['jobs']:
                jenkins_dict.setdefault(entry['name'], {'build_number': 277})
        jenkins = FakeJenkins(jenkins_dict, current_version=version)
        if not active:
            jenkins.set_build_result(BUILD_REVISION, 2, 'SUCCESS')
        return CIDirector(jenkins, state_file, branches)

    def test_schedule_builds_uses_get_active_jobs(self):
        director = CIDirector(FakeJenkins({
            REVISION_RESULTS: {'build_number': 2}
        }, current_version=5), StateFile())
        server_info = make_server_info('idle')
        director.state_file.start_revision('a', 54, 'b', '1.25', 5)
        # get_determine_result calls get_available_jobs, but doesn't count
        # because it doesn't run those jobs.
        with patch.object(
                ResultJudge, 'get_determine_result_jobs', return_value=[]):
            with patch.object(
                    director, 'get_active_jobs',
                    return_value=[BUILD_REVISION]) as gaj_mock:
                with patch.object(director.state_file, 'build_info_to_s3'):
                    director.schedule_builds(server_info)
        gaj_mock.assert_called_with(server_info)

    def test_schedule_builds_respects_branch_order(self):
        with bzr_sandbox() as bzr_home:
            foo_tree, foo_path, foo_rev = make_bzr_tree(bzr_home, 'foo', '1.7')
            bar_tree, bar_path, bar_rev = make_bzr_tree(bzr_home, 'bar',
                                                        '1.7', 2)
            director = self.make_schedule_builds_director([bar_path, foo_path])
            director.schedule_builds(make_server_info('idle'))
            self.assertEqual(
                director.jenkins.calls['build_job'],
                [((BUILD_REVISION, {
                    'branch': bar_path, 'revision': 2
                }), {})])
            director = self.make_schedule_builds_director([foo_path, bar_path])
            director.schedule_builds(make_server_info('idle'))
            self.assertEqual(
                director.jenkins.calls['build_job'],
                [((BUILD_REVISION, {
                    'branch': foo_path, 'revision': 1
                }), {})])

    def test_schedule_builds_updates_builds(self):
        director = self.make_schedule_builds_director([])
        with patch.object(director, 'update_builds') as ub_mock:
            director.schedule_builds(make_server_info('idle'))
        ub_mock.assert_called_with()

    def test_schedule_builds_records_results(self):
        director = self.make_schedule_builds_director([])
        with patch.object(RecordResultJob, 'maybe_build') as mb_mock:
            with patch.object(director.state_file, 'build_info_to_s3'):
                director.schedule_builds(make_server_info('idle'))
        mb_mock.assert_called_with()

    def test_schedule_builds_publishes(self):
        director = self.make_schedule_builds_director([], active=True)
        with patch.object(ResourcefulJob, 'maybe_build') as mb_mock:
            si = make_server_info('idle', buildable=[PUBLISH_REVISION])
            director.schedule_builds(si)
        mb_mock.assert_called_with(set())

    def test_schedule_builds_inactive_not_publishes(self):
        director = self.make_schedule_builds_director([], active=False)
        with patch.object(ResourcefulJob, 'maybe_build') as mb_mock:
            director.schedule_builds(make_server_info('idle'))
        self.assertEqual(mb_mock.call_count, 0)

    def test_schedule_builds_schedules_resourceful_tests_unique(self):
        server_info = make_server_info('idle', buildable=['package-foo'])
        director = self.make_schedule_builds_director(
            [], server_info=server_info)
        description = "[ci-director]\n requires: build-revision"
        director.jenkins.job_info['package-foo']['description'] = description
        with patch.object(ResourcefulJob, 'maybe_build') as mb_mock:
            director.schedule_builds(server_info)
        mb_mock.assert_called_once_with(set())

    def test_schedule_builds_adopts_builds(self):
        director = self.make_schedule_builds_director([])
        server_info = make_server_info('idle')
        with patch.object(director, 'adopt_builds') as ab_mock:
            director.schedule_builds(server_info)
        ab_mock.assert_called_with(server_info)

    def test_schedule_builds_calls_send_final_mail(self):
        director = self.make_schedule_builds_director([])
        server_info = make_server_info('idle')
        with patch.object(ResultJudge, 'finalize') as finalize_mock:
            with patch.object(director.state_file, 'build_info_to_s3'):
                director.schedule_builds(server_info)
        finalize_mock.assert_called_with(director.mailer)

    def test_schedule_builds_writes_results_to_s3_when_result_is_judged(self):
        director = self.make_schedule_builds_director([])
        with patch.object(ResultJudge, 'finalize', return_value=False):
            with patch.object(RecordResultJob, 'maybe_build',
                              return_value=True):
                with patch.object(director.state_file,
                                  'build_info_to_s3') as s3_mock:
                    director.schedule_builds(make_server_info('idle'))
        s3_mock.assert_called_with()

    def test_schedule_builds_writes_results_when_final_result_ready(self):
        director = self.make_schedule_builds_director([])
        with patch.object(ResultJudge, 'finalize', return_value=True):
            with patch.object(RecordResultJob, 'maybe_build',
                              return_value=False):
                with patch.object(director.state_file,
                                  'build_info_to_s3') as s3_mock:
                    director.schedule_builds(make_server_info('idle'))
        s3_mock.assert_called_with()

    def test_schedule_builds_no_write_results_to_s3_without_final_data(self):
        director = self.make_schedule_builds_director([])
        with patch.object(ResultJudge, 'finalize', return_value=False):
            with patch.object(RecordResultJob, 'maybe_build',
                              return_value=False):
                with patch.object(director.state_file,
                                  'build_info_to_s3') as s3_mock:
                    director.schedule_builds(make_server_info('idle'))
        self.assertEqual(0, s3_mock.call_count)

    def test_schedule_builds_results_to_s3_called_once_via_finalize(self):
        server_info = make_server_info('idle')
        director = self.make_schedule_builds_director(
            [], server_info=server_info)
        director.mailer = Mailer(None, None, None)
        judge = ResultJudge(None, director.state_file, director.jenkins)
        for job in judge.get_candidate_determine_result_jobs():
            finally_fail_job(self, job)
        with patch.object(director.state_file,
                          'build_info_to_s3') as s3_mock:
            with patch.object(director.mailer, 'send_mail'):
                with patch.object(RecordResultJob, 'maybe_build',
                                  return_value=False):
                    director.schedule_builds(server_info)
                    s3_mock.assert_called_once_with()
                    director.schedule_builds(server_info)
                    s3_mock.assert_called_once_with()

    def test_schedule_builds_results_to_s3_called_via_revision_res_job(self):
        server_info = make_server_info('idle')
        director = self.make_schedule_builds_director(
            [], server_info=server_info)
        director.mailer = Mailer(None, None, None)
        judge = ResultJudge(None, director.state_file, director.jenkins)
        for job in judge.get_candidate_determine_result_jobs():
            if job.test_id != 'local-deploy-trusty-arm64-devel':
                finally_fail_job(self, job)
        with patch.object(director.state_file,
                          'build_info_to_s3') as s3_mock:
            with patch.object(director.mailer, 'send_mail'):
                with patch.object(ResultJudge, 'finalize',
                                  return_value=False):
                    director.schedule_builds(server_info)
                    s3_mock.assert_called_once_with()
                    director.schedule_builds(server_info)
                    s3_mock.assert_called_once_with()

    def test_from_config(self):
        director = CIDirector.from_config({
            'jenkins_url': 'http://192.168.1.1:8080',
            'jenkins_user': 'jrandom',
            'jenkins_password': 'jrpass',
            'from_address': 'from@example.com',
            'to_address': 'to@example.com',
            'site_url': 'site:foo',
        }, StateFile(), [])
        self.assertEqual(director.jenkins.server, 'http://192.168.1.1:8080/')
        self.assertEqual(director.jenkins.auth,
                         auth_headers('jrandom', 'jrpass'))

    def test_current_version_published_no_current_version(self):
        director = CIDirector(FakeJenkins({}, current_version=55), StateFile())
        self.assertEqual(False, director.current_version_published())

    def test_current_version_published_ready(self):
        director = CIDirector(FakeJenkins({}, current_version=55), StateFile())
        director.state_file.start_revision('a', 54, 'b', '1.10', 55)
        pub_job = director.state_file.publication_job()
        pub_job.update_from_build_result('SUCCESS')
        self.assertTrue(director.current_version_published())

    def test_current_version_published_failure(self):
        director = CIDirector(FakeJenkins({}, current_version=55), StateFile())
        director.state_file.start_revision('a', 54, 'b', '1.7', 55)
        pub_job = director.state_file.publication_job()
        pub_job.start_build(5)
        finally_fail_job(self, pub_job)
        self.assertIs(False, director.current_version_published())

    def test_current_version_published_stale(self):
        director = CIDirector(FakeJenkins({}, current_version=56), StateFile())
        director.state_file.start_revision('a', 54, 'b', '1.8', 55)
        pub_job = director.state_file.publication_job()
        pub_job.update_from_build_result('SUCCESS')
        self.assertEqual(True, director.current_version_published())

    def test_update_builds_no_version(self):
        sf = StateFile()
        jenkins = FakeJenkins({})
        director = CIDirector(jenkins, sf)
        director.update_builds()
        self.assertEqual({'get_info': [((), {})]}, jenkins.calls)

    def test_update_builds_no_builds(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        sf.get_build_revision_job().update_from_build_result('SUCCESS')
        jenkins = FakeJenkins({})
        director = CIDirector(jenkins, sf)
        director.update_builds()
        self.assertEqual({'get_info': [((), {})]}, jenkins.calls)

    def test_update_builds_with_test(self):
        sf = self.get_state_file()
        sf.get_grouped_job('test-id').start_build(5)
        jenkins = FakeJenkins({'test-id': {'build_number': 5}},
                              current_version=sf.current_version_number)
        director = CIDirector(jenkins, sf)
        director.update_builds()
        self.assertEqual({
            'get_build_info': [(('test-id', 5), {})],
            'get_info': [((), {})],
        }, jenkins.calls)

    def test_update_builds_with_building_test(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        sf.get_build_revision_job().update_from_build_result('SUCCESS')
        sf.get_grouped_job('test-id').start_build(5)
        jenkins = FakeJenkins({
            'test-id': {'build_number': 6, 'building': True}
        })
        director = CIDirector(jenkins, sf)
        director.update_builds()
        self.assertEqual({'get_info': [((), {})]}, jenkins.calls)

    def test_update_builds_with_building_resourceful_jobs(self):
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20.1', 2)
        sf.get_build_revision_job().update_from_build_result('SUCCESS')
        sf.get_grouped_job(PUBLISH_REVISION, 'release-tests').start_build(5)
        sf.get_grouped_job('foo-deploy', 'functional-tests')
        jenkins = FakeJenkins({
            PUBLISH_REVISION: {'build_number': 6, 'building': True},
            'foo-deploy': {
                'build_number': 6,
                'building': False,
                'description': '[ci-director]\n requires: build-revision'},
        })
        director = CIDirector(jenkins, sf)
        director.update_builds()
        job_ids = [j.test_id for j in sf.iter_grouped_jobs()]
        self.assertEqual([PUBLISH_REVISION, 'foo-deploy'], job_ids)

    def get_state_file(self, published=True, built=True):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        if built:
            sf.get_build_revision_job().update_from_build_result('SUCCESS')
        if published:
            sf.publication_job().update_from_build_result('SUCCESS')
        return sf

    def test_update_builds_updates_from_status(self):
        sf = self.get_state_file()
        sf.get_grouped_job('test-id').start_build(6)
        jenkins = FakeJenkins({
            'test-id': {'build_number': 6}
        })
        jenkins.build_info['test-id'][6]['result'] = 'RANDOM'
        director = CIDirector(jenkins, sf)
        with patch.object(StateFileJob, 'update_from_build_result') as mb:
            director.update_builds()
        mb.assert_called_with('RANDOM')

    def test_update_builds_includes_publication(self):
        sf = self.get_state_file(published=False)
        pub_job = sf.publication_job()
        pub_job.start_build(23)
        jenkins = FakeJenkins({PUBLISH_REVISION: {'build_number': 23}})
        director = CIDirector(jenkins, sf)
        jenkins.set_build_result(PUBLISH_REVISION, 23, None)
        director.update_builds()
        jenkins.set_build_result(PUBLISH_REVISION, 23, 'SUCCESS')
        self.assertEqual(BUILDING, pub_job.get_status())
        director.update_builds()
        self.assertEqual(SUCCEEDED, pub_job.get_status())

    def set_build_result(self, job, jenkins, result):
        jenkins.set_build_result(job.test_id, job.build_number, result)

    def test_update_builds_includes_build_revision(self):
        sf = self.get_state_file(built=False)
        sf_build_revision = sf.get_build_revision_job()
        self.assertEqual(sf_build_revision.get_status(), BUILDING)
        jenkins = FakeJenkins({})
        self.set_build_result(sf_build_revision, jenkins, 'SUCCESS')
        director = CIDirector(jenkins, sf)
        director.update_builds()
        self.assertEqual(sf_build_revision.get_status(), SUCCEEDED)

    def test_update_builds_failed_test(self):
        sf = self.get_state_file()
        sfj = sf.get_grouped_job('foo-deploy')
        finally_fail_job(self, sfj)
        self.assertEqual(sfj.get_status(), FAILED)
        sfj.start_build(27)
        jenkins = FakeJenkins({})
        self.set_build_result(sfj, jenkins, 'SUCCESS')
        director = CIDirector(jenkins, sf)
        director.update_builds()
        self.assertEqual(sfj.get_status(), SUCCEEDED)

    def test_adopt_builds(self):
        jenkins = FakeJenkins({
            'foo-deploy': {
                'build_number': 5,
                'description': '[ci-director]\n requires: build-revision'},
            BuildRevisionJob.job_id: {'build_number': 2},
        })
        sf = self.get_state_file()
        director = CIDirector(jenkins, sf)
        with patch.object(ResourcefulJob, 'adopt_build') as ab_mock:
            director.adopt_builds(make_server_info('idle'))
            self.assertEqual(ab_mock.call_count, 0)
            director.adopt_builds(
                make_server_info(
                    'idle', buildable=['foo-deploy']))
        ab_mock.assert_called_with({
            'result': 'SUCCESS',
            'number': 5,
            'building': False,
            'artifacts': [],
            'url': 'fakettp://fake.fake/job/foo-deploy/5'
        })

    def test_adopt_builds_unbuilt(self):
        info = ServerInfo({'jobs': [{
            'name': 'aws-deploy',
        }]})
        jenkins = FakeJenkins({
            BuildRevisionJob.job_id: {'build_number': 2},
            'aws-deploy': {
                'build_number': 2,
                'description': '[ci-director]\n requires: build-revision',
                'lastBuild': None}
        })
        jenkins.job_info['aws-deploy'] = {'lastBuild': None}
        sf = StateFile()
        director = CIDirector(jenkins, sf)
        director.adopt_builds(info)
        self.assertEqual(
            [c[0][0] for c in jenkins.calls['get_build_info']],
            ['build-revision'])
        jenkins.job_info['aws-deploy'] = {
            'lastBuild': {'number': 1},
            'description': '[ci-director]\n requires: build-revision'}
        jenkins.build_info['aws-deploy'] = {1: {}}
        director.adopt_builds(info)
        self.assertEqual(
            [c[0][0] for c in jenkins.calls['get_build_info']],
            ['build-revision', 'build-revision', 'aws-deploy'])


def finally_fail(test_case, sf, test_id):
    sf.get_grouped_job(test_id).start_build(5)
    finally_fail_job(test_case, sf.get_grouped_job(test_id))


class TestMailer(TestCase):

    def test_from_config(self):
        config = {
            'to_address': 'to1@example.com',
            'from_address': 'from1@example.com',
            'site_url': 'site:foo1',
            'bar': 'baz',
        }
        mailer = Mailer.from_config(config)
        self.assertEqual(mailer.to_address, 'to1@example.com')
        self.assertEqual(mailer.from_address, 'from1@example.com')
        self.assertEqual(mailer.site_url, 'site:foo1')

    def test_format_failure(self):
        summary = {
            'branch': 'lp:foo/bar',
            'revno': 15,
            'revision_id': 'unused',
            'short_revision': 1234567890,
            'build_number': 2,
            'release_number': '1.6',
        }
        mailer = Mailer(None, None, 'site:foo')
        output = mailer.format_failure(
            summary, [('qa', 5), ('ab', 6)], [])
        self.assertEqual(output, dedent("""\
            Build: #2 Revision: lp:foo/bar 1234567890 Version: 1.6

            Failed tests
            ab build #6 site:foo/job/ab/6/console
            qa build #5 site:foo/job/qa/5/console
        """))

    def test_format_failure_non_voting_only(self):
        summary = {
            'branch': 'lp:foo/bar',
            'revno': 15,
            'revision_id': 'unused',
            'short_revision': 1234567890,
            'build_number': 2,
            'release_number': '1.6',
        }
        mailer = Mailer(None, None, 'site:foo')
        output = mailer.format_failure(
            summary, [], [('qa', 5), ('ab', 6)])
        self.assertEqual(output, dedent("""\
            Build: #2 Revision: lp:foo/bar 1234567890 Version: 1.6

            Failed non-voting tests
            ab build #6 site:foo/job/ab/6/console
            qa build #5 site:foo/job/qa/5/console
        """))

    def test_format_failure_voting_and_non_voting_only(self):
        summary = {
            'branch': 'lp:foo/bar',
            'revno': 15,
            'revision_id': 'unused',
            'short_revision': 1234567890,
            'build_number': 2,
            'release_number': '1.6',
        }
        mailer = Mailer(None, None, 'site:foo')
        output = mailer.format_failure(
            summary, [('qa', 5), ('ab', 6)], [('foo', 7), ('bar', 8)])
        self.assertEqual(output, dedent("""\
            Build: #2 Revision: lp:foo/bar 1234567890 Version: 1.6

            Failed tests
            ab build #6 site:foo/job/ab/6/console
            qa build #5 site:foo/job/qa/5/console
            Failed non-voting tests
            bar build #8 site:foo/job/bar/8/console
            foo build #7 site:foo/job/foo/7/console
        """))

    def test_get_failure_message(self):
        summary = {
            'branch': 'lp:foo/bar',
            'revno': 15,
            'short_revision': 12345678,
            'revision_id': 'unused',
            'build_number': 2,
            'release_number': '1.6',
        }
        mailer = Mailer('from@example.com', 'to@example.com', 'site:foo')
        message = mailer.get_failure_message(
            summary, iter([('qa', 5), ('ab', 6)]), [])
        self.assertEqual(dedent("""\
            To: to@example.com
            From: CI & CD Jenkins <from@example.com>
            Subject: Cursed: #2 lp:foo/bar 12345678 (ab, qa)

            Build: #2 Revision: lp:foo/bar 12345678 Version: 1.6

            Failed tests
            ab build #6 site:foo/job/ab/6/console
            qa build #5 site:foo/job/qa/5/console
        """), message.as_string())

    def test_get_failure_message_non_voting_only(self):
        summary = {
            'branch': 'lp:foo/bar',
            'revno': 15,
            'short_revision': 12345678,
            'revision_id': 'unused',
            'build_number': 2,
            'release_number': '1.6',
        }
        mailer = Mailer('from@example.com', 'to@example.com', 'site:foo')
        message = mailer.get_failure_message(
            summary, [], iter([('ab', 1), ('cd', 2)]))
        self.assertEqual(dedent("""\
            To: to@example.com
            From: CI & CD Jenkins <from@example.com>
            Subject: Info: #2 lp:foo/bar 12345678 ()

            Build: #2 Revision: lp:foo/bar 12345678 Version: 1.6

            Failed non-voting tests
            ab build #1 site:foo/job/ab/1/console
            cd build #2 site:foo/job/cd/2/console
        """), message.as_string())

    def test_get_failure_message_final(self):
        summary = {
            'branch': 'lp:foo/bar',
            'revno': 15,
            'short_revision': 'r15',
            'revision_id': 'unused',
            'build_number': 2,
            'release_number': '1.6',
        }
        mailer = Mailer('from@example.com', 'to@example.com', 'site:foo')
        message = mailer.get_failure_message(
            summary, iter([('qa', 5), ('ab', 6)]), [], final=True)
        self.assertEqual(dedent("""\
            To: to@example.com
            From: CI & CD Jenkins <from@example.com>
            Subject: Cursed (final): #2 lp:foo/bar r15 (ab, qa)

            Build: #2 Revision: lp:foo/bar r15 Version: 1.6

            Failed tests
            ab build #6 site:foo/job/ab/6/console
            qa build #5 site:foo/job/qa/5/console
        """), message.as_string())


class TestResultJudge(TestCase):

    def test_determine_result_pending(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        jenkins = FakeJenkins({
            'aws-deploy': {
                'build_number': 20,
                'description': '[ci-director]\n requires: build-revision',
            },
            'aws-upgrade': {
                'build_number': 20,
                'description': '[ci-director]\n requires: build-revision',
            },
        })
        judge = ResultJudge(logging.getLogger(), sf, jenkins)
        self.assertEqual(PENDING, judge.determine_result())
        sf.get_grouped_job('aws-deploy').start_build(5)
        sfj = sf.get_grouped_job('aws-deploy')
        sfj.update_from_build_result('SUCCESS')
        self.assertEqual(PENDING, judge.determine_result())

    def test_determine_result_curse(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        jenkins = FakeJenkins({
            'aws-deploy': {
                'build_number': 20,
                'description': '[ci-director]\n requires: build-revision',
            },
            'aws-upgrade': {
                'build_number': 20,
                'description': '[ci-director]\n requires: build-revision',
            },
        })
        job = ResultJudge(logging.getLogger(), sf, jenkins)
        self.assertEqual(PENDING, job.determine_result())
        finally_fail(self, sf, 'aws-upgrade')
        self.assertEqual(CURSE, job.determine_result())

    def test_determine_result_bless(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        build_revision = sf.get_build_revision_job()
        build_revision.update_from_build_result('SUCCESS')
        jenkins = FakeJenkins({
            'aws-deploy': {
                'build_number': 20,
                'description': '[ci-director]\n requires: build-revision',
            }
        })
        sf.get_grouped_job('aws-deploy').start_build(5)
        sfj = sf.get_grouped_job('aws-deploy')
        sfj.update_from_build_result('SUCCESS')
        judge = ResultJudge(logging.getLogger(), sf, jenkins)
        self.assertEqual(BLESS, judge.determine_result())

    def make_judge(self, available_jobs=None, disabled_jobs=None):
        if available_jobs is None:
            available_jobs = ['aws-deploy', 'mws-upgrade']
        jobs = dict(
            (j, {'build_number': 20}) for j in available_jobs)
        if disabled_jobs is not None:
            jobs.update(
                (j, {'color': 'disabled', 'build_number': 20})
                for j in disabled_jobs)
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        jenkins = FakeJenkins(jobs)
        return ResultJudge(logging.getLogger(), sf, jenkins)

    def test_get_candidate_determine_result_jobs_resoureful_jobs(self):
        judge = self.make_judge(['aws-deploy', 'functional-foo'])
        description = (
            "[ci-director]\nrequires: aws-deploy\ngroup-name: functional-tests"
        )
        judge.jenkins.job_info['aws-deploy']['description'] = description
        judge.jenkins.job_info['functional-foo']['description'] = description
        expected = ['aws-deploy', 'build-revision', 'functional-foo']
        result = [
            sfj.test_id for sfj in judge.get_candidate_determine_result_jobs()]
        self.assertEqual(expected, sorted(result))

    def test_determine_result_ignores_non_tests(self):
        judge = self.make_judge()
        state_file = judge.state_file
        build_revision = state_file.get_build_revision_job()
        build_revision.update_from_build_result('SUCCESS')
        state_file.get_grouped_job('aws-deploy').start_build(5)
        sfj = state_file.get_grouped_job('aws-deploy')
        sfj.update_from_build_result('SUCCESS')
        self.assertEqual(BLESS, judge.determine_result())

    def test_determine_result_includes_build(self):
        judge = self.make_judge()
        self.assertEqual(judge.determine_result(), PENDING)
        build_revision = judge.state_file.get_build_revision_job()
        build_revision.update_from_build_result('FAILURE')
        self.assertEqual(judge.determine_result(), CURSE)

    def test_determine_result_includes_resourceful_jobs(self):
        judge = self.make_judge(['aws-deploy', 'functional-foo'])
        description = (
            "[ci-director]\nrequires: aws-deploy\ngroup-name: functional-tests"
        )
        judge.jenkins.job_info['functional-foo']['description'] = description
        state_file = judge.state_file
        state_file.get_grouped_job('aws-deploy').start_build(5)
        build_revision = state_file.get_build_revision_job()
        build_revision.update_from_build_result('SUCCESS')
        awsj = state_file.get_grouped_job('aws-deploy')
        awsj.update_from_build_result('SUCCESS')
        self.assertEqual(PENDING, judge.determine_result())
        fooj = state_file.get_grouped_job(
            'functional-foo', 'functional-tests', None)
        fooj.update_from_build_result('SUCCESS')
        self.assertEqual(BLESS, judge.determine_result())
        fooj._data['status'] = PENDING
        finally_fail_job(self, fooj)
        self.assertEqual(CURSE, judge.determine_result())

    def test_determine_result_excludes_disabled(self):
        judge = self.make_judge(available_jobs=[],
                                disabled_jobs=['aws-deploy'])
        state_file = judge.state_file
        build_revision = state_file.get_build_revision_job()
        build_revision.update_from_build_result('SUCCESS')
        self.assertEqual(judge.determine_result(), BLESS)

    def test_determine_result_ignores_vote_false(self):
        devel_job = StateFileJob(
            'job-no_vote', {'failure_count': 1}, vote=False)
        job = StateFileJob('job', {'status': SUCCEEDED})
        judge = self.make_judge(available_jobs=['job-devel', 'job'])
        with patch.object(
                judge, 'get_candidate_determine_result_jobs',
                return_value=[job, devel_job]):
            self.assertEqual(BLESS, judge.determine_result())

    def test_list_failures(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        directives = '[ci-director]\n failure-threshold: 3\n vote: {}'
        jenkins = FakeJenkins({
            'aws-deploy': {
                'build_number': 2, 'description': directives.format(True)},
            'aws-no_vote': {
                'build_number': 2, 'description': directives.format(False)},
        })
        si = make_server_info('idle', buildable=['aws-deploy', 'aws-no_vote'])
        for job in ResourcefulJob.get_available_jobs(si, jenkins, sf, None):
            job.get_state_file_job()
        # Move the jobs though all the possible states without reaching
        # the failure-threshold to show there is nothing to judge.
        judge = ResultJudge(None, sf, jenkins)
        for result in JENKINS_BUILD_RESULTS:
            sf.get_grouped_job(
                'aws-deploy').update_from_build_result(result)
            sf.get_grouped_job(
                'aws-no_vote').update_from_build_result(result)
            voting_failures, non_voting_failures = judge.list_failures()
            self.assertItemsEqual([], voting_failures)
            self.assertItemsEqual([], non_voting_failures)
        # Complete all retries of the jobs.
        sf.get_grouped_job('aws-deploy')._data['status'] = PENDING
        finally_fail_job(self, sf.get_grouped_job('aws-deploy'), 72)
        sf.get_grouped_job('aws-no_vote')._data['status'] = PENDING
        finally_fail_job(self, sf.get_grouped_job('aws-no_vote'), 72)
        voting_failures, non_voting_failures = judge.list_failures()
        self.assertItemsEqual(
            voting_failures, [('aws-deploy', 72)])
        self.assertItemsEqual(
            non_voting_failures, [('aws-no_vote', 72)])

    def test_result_tests_complete(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        jenkins = FakeJenkins({})
        judge = ResultJudge(None, sf, jenkins)
        for job in judge.get_determine_result_jobs():
            # Not complete before all results are in.
            self.assertEqual(False, judge.result_tests_complete())
            finally_fail_job(self, job)
        # Complete when all determine_result_jobs are failed.
        self.assertEqual(True, judge.result_tests_complete())
        # Successes are okay, too.
        for job in judge.get_determine_result_jobs():
            job.update_from_build_result('SUCCESS')
            self.assertEqual(True, judge.result_tests_complete())

    def test_send_final_mail(self):
        judge = ResultJudge(None, StateFile(), FakeJenkins({}))
        judge.state_file.start_revision('a', 1, 'b', '2.2', 3)
        for job in judge.get_determine_result_jobs():
            finally_fail_job(self, job)
        mailer = Mailer('from', 'to', 'url')
        sentry = object()
        with patch.object(mailer, 'get_failure_message',
                          return_value=sentry) as gfm:
            with patch.object(mailer, 'send_mail') as sm_mock:
                judge.finalize(mailer)
        gfm.assert_called_with(
            judge.state_file.get_current_version_summary(),
            [('build-revision', 3)], [], final=True)
        sm_mock.assert_called_with(sentry)

    def test_send_final_requires_complete(self):
        judge = ResultJudge(None, StateFile(), FakeJenkins({}))
        judge.state_file.start_revision('a', 1, 'b', '2.2', 3)
        finally_fail_job(self, judge.state_file.get_build_revision_job())
        mailer = Mailer('from', 'to', 'url')
        with patch.object(mailer, 'get_failure_message',
                          return_value=None) as gfm:
            with patch.object(mailer, 'send_mail'):
                with patch.object(judge, 'result_tests_complete',
                                  return_value=False):
                    self.assertFalse(judge.finalize(mailer))
                    self.assertEqual(0, gfm.call_count)
                    self.assertIs(
                        None, judge.state_file.get_complete_failure_info(3))
                with patch.object(judge, 'result_tests_complete',
                                  return_value=True):
                    self.assertTrue(judge.finalize(mailer))
                    self.assertEqual(1, gfm.call_count)
                    self.assertEqual(
                        {'build-revision': 3},
                        judge.state_file.get_complete_failure_info(3))

    def test_send_final_mail_requires_nonmatching_failure_info(self):
        judge = ResultJudge(None, StateFile(), FakeJenkins({}))
        judge.state_file.start_revision('a', 1, 'b', '2.2', 3)
        for job in judge.get_determine_result_jobs():
            finally_fail_job(self, job)
        judge.state_file.set_complete_failure_info(
            3, dict(judge.list_failures()[0]))
        mailer = Mailer('from', 'to', 'url')
        with patch.object(mailer, 'get_failure_message') as gfm:
            with patch.object(mailer, 'send_mail') as sm_mock:
                judge.finalize(mailer)
                self.assertEqual(0, gfm.call_count)
                self.assertEqual(0, sm_mock.call_count)
                judge.state_file.get_complete_failure_info(3)['foobar'] = 4
                judge.finalize(mailer)
        self.assertEqual(1, gfm.call_count)
        self.assertEqual(1, sm_mock.call_count)

    def test_send_final_mail_skips_on_success(self):
        judge = ResultJudge(None, StateFile(), FakeJenkins({}))
        judge.state_file.start_revision('a', 1, 'b', '2.2', 3)
        for job in judge.get_determine_result_jobs():
            job.update_from_build_result('SUCCESS')
        mailer = Mailer('from', 'to', 'url')
        with patch.object(mailer, 'get_failure_message') as gfm:
            with patch.object(mailer, 'send_mail') as sm_mock:
                self.assertTrue(judge.finalize(mailer))
                self.assertEqual(0, gfm.call_count)
                self.assertEqual(0, sm_mock.call_count)
                for job in judge.get_determine_result_jobs():
                    job._data['status'] = PENDING
                    finally_fail_job(self, job)
                self.assertTrue(judge.finalize(mailer))
        self.assertEqual(1, gfm.call_count)
        self.assertEqual(1, sm_mock.call_count)

    def test_send_final_mail_updates_after_sent(self):
        judge = ResultJudge(None, StateFile(), FakeJenkins({}))
        judge.state_file.start_revision('a', 1, 'b', '2.2', 3)
        for job in judge.get_determine_result_jobs():
            finally_fail_job(self, job)
        mailer = Mailer('from', 'to', 'url')
        with patch.object(mailer, 'get_failure_message'):
            with patch.object(mailer, 'send_mail', side_effect=ValueError):
                try:
                    judge.finalize(mailer)
                except ValueError:
                    pass
        self.assertIs(None, judge.state_file.get_complete_failure_info(3))

    def test_finalize_indicates_final_status_ready_all_tests_succeed(self):
        jenkins = FakeJenkins({
            'aws-deploy-devel': {
                'build_number': 2,
                'description': '[ci-director]\n vote: false'},
        })
        judge = ResultJudge(None, StateFile(), jenkins)
        judge.state_file.start_revision('a', 1, 'b', '2.2', 3)
        mailer = Mailer('from', 'to', 'url')
        # Tests are still running
        self.assertFalse(judge.finalize(mailer))
        for job in judge.get_determine_result_jobs():
            job.update_from_build_result('SUCCESS')
        # The -devel job is still running
        self.assertFalse(judge.finalize(mailer))
        for job in judge.get_candidate_determine_result_jobs():
            if not job.vote:
                job.update_from_build_result('SUCCESS')
        # The first call of finalize() when all jobs are finished
        # indicates that all data is now available in the state file.
        self.assertTrue(judge.finalize(mailer))
        # Another call of finalize() returns False because
        # nothing has changed.
        self.assertFalse(judge.finalize(mailer))

    def test_finalize_indicates_final_status_important_real_test_fails(self):
        jenkins = FakeJenkins({
            'aws-deploy-devel': {
                'build_number': 2,
                'description': '[ci-director]\n vote: false'},
        })
        judge = ResultJudge(None, StateFile(), jenkins)
        judge.state_file.start_revision('a', 1, 'b', '2.2', 3)
        mailer = Mailer('from', 'to', 'url')
        with patch.object(mailer, 'get_failure_message'):
            with patch.object(mailer, 'send_mail'):
                # Tests are still running
                self.assertFalse(judge.finalize(mailer))
                for job in judge.get_determine_result_jobs():
                    finally_fail_job(self, job)
                # The -devel job is still running
                self.assertFalse(judge.finalize(mailer))
                for job in judge.get_candidate_determine_result_jobs():
                    if not job.vote:
                        job.update_from_build_result('SUCCESS')
                # The first call of finalize() when all jobs are finished
                # indicates that all data is now available in the state file.
                self.assertTrue(judge.finalize(mailer))
                # Another call of finalize() returns False because
                # nothing has changed.
                self.assertFalse(judge.finalize(mailer))


class TestRecordResultJob(TestCase):

    def test_build_get_number_none(self):
        jenkins = FakeJenkins({
            'foo': {'build_number': None}
        })

        class TestJob(RecordResultJob):
            job_id = 'foo'
        results = iter([
            {'lastBuild': None}, {'lastBuild': None},
            {'lastBuild': {'number': 1}}])

        def results_get(job_id):
            return results.next()

        job = TestJob(logging.getLogger(), StateFile(), jenkins)
        with patch.object(jenkins, 'get_job_info',
                          side_effect=results_get):
            job.build_get_number()

    def test_maybe_build_no_revision(self):
        jenkins = FakeJenkins({})
        sf = StateFile()
        job = RecordResultJob(logging.getLogger(), sf, jenkins)
        with patch.object(job.judge, 'determine_result',
                          return_value=BLESS) as dr_mock:
            with patch.object(job, 'build_get_number') as bn_mock:
                self.assertEqual(False, job.maybe_build())
        self.assertEqual(0, dr_mock.call_count)
        self.assertEqual(0, bn_mock.call_count)

    def test_maybe_build_pending(self):
        jenkins = FakeJenkins({})
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        job = RecordResultJob(logging.getLogger(), sf, jenkins)
        with patch.object(job.judge, 'determine_result',
                          return_value=PENDING) as dr_mock:
            with patch.object(job, 'build_get_number') as bn_mock:
                self.assertEqual(False, job.maybe_build())
        dr_mock.assert_called_with()
        self.assertEqual(0, bn_mock.call_count)
        self.assertIs(sf.get_result()[0], None)

    def test_maybe_build_bless(self):
        jenkins = FakeJenkins({})
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        job = RecordResultJob(logging.getLogger(), sf, jenkins)
        with patch.object(job.judge, 'determine_result', return_value=BLESS
                          ) as dr_mock:
            with patch.object(job, 'build_get_number',
                              return_value=23) as bn_mock:
                self.assertEqual(True, job.maybe_build())
        dr_mock.assert_called_with()
        bn_mock.assert_called_with({
            'build_number': 2,
            'blessed': 'true',
            'revision_id': 'b',
            'revno': 1,
            'short_revision': 'r1',
            'build_number': 2,
            'branch': 'a',
            'release_number': '1.6',
        })
        self.assertEqual(sf.get_result(), (BLESS, 23))

    def test_maybe_build_curse(self):
        jenkins = FakeJenkins({})
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        mailer = Mailer('from@example.com', 'to@example.com', 'site:foo')
        job = RecordResultJob(logging.getLogger(), sf, jenkins, mailer)
        with patch.object(job.judge, 'determine_result', return_value=CURSE
                          ) as dr_mock:
            with patch.object(job, 'build_get_number',
                              return_value=23) as bn_mock:
                with patch('smtplib.SMTP.sendmail'):
                    with patch('smtplib.SMTP.connect'):
                        self.assertEqual(True, job.maybe_build())
        dr_mock.assert_called_with()
        bn_mock.assert_called_with({
            'build_number': 2,
            'blessed': 'false',
            'revision_id': 'b',
            'revno': 1,
            'short_revision': 'r1',
            'build_number': 2,
            'branch': 'a',
            'release_number': '1.6',
        })
        self.assertEqual(sf.get_result(), (CURSE, 23))

    def test_maybe_build_curse_sends_mail(self):
        jenkins = FakeJenkins({REVISION_RESULTS: {'build_number': 5}})
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        finally_fail_job(self, sf.get_build_revision_job())
        mailer = Mailer('from@example.com', 'to@example.com', 'site:foo')
        job = RecordResultJob(logging.getLogger(), sf, jenkins, mailer)
        with patch('smtplib.SMTP.connect') as connect_mock:
            with patch('smtplib.SMTP.sendmail') as sm_mock:
                self.assertTrue(job.maybe_build())
        connect_mock.assert_called_with()
        summary = sf.get_current_version_summary()
        message = mailer.get_failure_message(
            summary, [('build-revision', 2)], [])
        sm_mock.assert_called_with(
            mailer.from_address, mailer.to_address, message.as_string())

    def test_maybe_build_bless_does_not_send_mail(self):
        jenkins = FakeJenkins({REVISION_RESULTS: {'build_number': 5}})
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        sf.get_build_revision_job().update_from_build_result('SUCCESS')
        job = RecordResultJob(logging.getLogger(), sf, jenkins)
        with patch('smtplib.SMTP.sendmail') as sm_mock:
            self.assertTrue(job.maybe_build())
        self.assertEqual(0, sm_mock.call_count)

    def test_no_build_if_same(self):
        jenkins = FakeJenkins({})
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        sf.save_result(CURSE, 12)
        job = RecordResultJob(logging.getLogger(), sf, jenkins)
        with patch.object(job.judge, 'determine_result', return_value=CURSE
                          ) as dr_mock:
            with patch.object(job, 'build_get_number',
                              return_value=23) as bn_mock:
                self.assertEqual(False, job.maybe_build())
        dr_mock.assert_called_with()
        self.assertEqual(0, bn_mock.call_count)
        self.assertEqual(sf.get_result(), (CURSE, 12))


class TestServerInfo(TestCase):

    def test_get_building_jobs(self):
        si = ServerInfo({'jobs': []})
        self.assertEqual([], si.get_building_jobs())
        si = ServerInfo({'jobs': [
            {'color': 'blue', 'name': 'a'},
            {'color': 'red', 'name': 'b'},
        ]})
        self.assertEqual([], si.get_building_jobs())
        si = ServerInfo({'jobs': [
            {'color': 'blue', 'name': 'a'},
            {'color': 'red', 'name': 'b'},
            {'color': 'blue_anime', 'name': 'c'},
            {'color': 'red_anime', 'name': 'd'},
        ]})
        self.assertEqual(['c', 'd'], si.get_building_jobs())

    def test_get_enabled_jobs(self):
        si = make_server_info(
            'idle', buildable=['foo', 'bar'], building=['baz'],
            disabled=['qux'])
        self.assertEqual(si.get_enabled_jobs(), ['foo', 'bar', 'baz'])


class TestConfigReader(TestCase):

    def test_config_file_name(self):
        with patch.dict(os.environ, {'HOME': '/home/jrandomx'}):
            self.assertEqual('/home/jrandomx/.config/ci-director',
                             ConfigReader.config_file_name())

    def test_read_config(self):
        with temp_dir() as home_dir:
            os.mkdir(os.path.join(home_dir, '.config'))
            with patch.dict(os.environ, {'HOME': home_dir}):
                config_file_name = ConfigReader.config_file_name()
                with open(config_file_name, 'w') as config_file:
                    config_file.write((dedent("""\
                    foo: bar
                    asdf: baz
                    """)))
                self.assertEqual(ConfigReader.read_config(), {
                    'foo': 'bar',
                    'asdf': 'baz',
                })


class JobSubclass(Job):

    job_id = 'job-subclass'

    def get_build_parameters(self):
        """Return the parameters for building, as a dict."""
        return None

    def get_prerequisite_jobs(self):
        return [self.state_file.get_build_revision_job(),
                self.state_file.publication_job()]

    def get_state_file_job(self):
        data = self.state_file._current_version
        job_data = data.setdefault('job-subclass', {'failures': 0})
        return StateFileJob('job-subclass', job_data)


def make_build_info(number, revision_build=None, branch=None, revision=None,
                    building=True):
    param_dict = {}
    if revision_build is not None:
        param_dict['revision_build'] = str(revision_build)
    if branch is not None:
        param_dict['branch'] = branch
    if revision is not None:
        param_dict['revision'] = str(revision)
    parameters = [{'name': key, 'value': value} for key, value in
                  param_dict.items()]
    return {
        'number': number,
        'actions': [{'parameters': parameters}],
        'building': building,
    }


class TestJob(TestCase):

    def test_adopt_build_requires_matching_revision_build(self):
        jenkins = FakeJenkins({})
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.2.4', 1065)
        build_info = {'number': 142}
        job = JobSubclass(logging.getLogger(), sf, jenkins)
        with patch.object(StateFileJob, 'start_build') as sb_mock:
            job.adopt_build(build_info)
            self.assertEqual(sb_mock.call_count, 0)
            build_info['actions'] = [{'parameters': [{
                'name': 'revision_build', 'value': '1066'
            }]}]
            job.adopt_build(build_info)
            self.assertEqual(sb_mock.call_count, 0)
            build_info['actions'] = [{'parameters': [{
                'name': 'revision_build', 'value': '1065'
            }]}]
            job.adopt_build(build_info)
            sb_mock.assert_called_with((142))

    def test_adopt_build_requires_newer_build(self):
        jenkins = FakeJenkins({})
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.2.4', 1065)
        job = JobSubclass(logging.getLogger(), sf, jenkins)
        job.get_state_file_job().start_build(25)
        with patch.object(StateFileJob, 'start_build') as sb_mock:
            job.adopt_build(make_build_info(24, 1065))
            self.assertEqual(sb_mock.call_count, 0)
            job.adopt_build(make_build_info(25, 1065))
            self.assertEqual(sb_mock.call_count, 0)
            job.adopt_build(make_build_info(26, 1065))
            self.assertEqual(sb_mock.call_count, 1)

    def test_adopt_build_works_if_no_build(self):
        jenkins = FakeJenkins({})
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.2.4', 1065)
        job = JobSubclass(logging.getLogger(), sf, jenkins)
        with patch.object(StateFileJob, 'start_build') as sb_mock:
            job.adopt_build(make_build_info(24, 1065))
            self.assertEqual(sb_mock.call_count, 1)

    def test_make_subclass(self):
        jenkins = FakeJenkins({
            'foo': {
                'build_number': 7,
                'description': '[ci-director]\n requires: build-revision'},
        })
        state_file = StateFile()
        logger = logging.getLogger()
        instance = Job.make_subclass('foo', logger, state_file, jenkins)
        self.assertIsInstance(instance, ResourcefulJob)
        self.assertEqual('foo', instance.job_id)

    def make_job(self, prerequisites=True):
        state_file = StateFile()
        state_file.start_revision('a', 5, 'b', '1.23.5', 29)
        jenkins = FakeJenkins({JobSubclass.job_id: {'build_number': 11}})
        job = JobSubclass(
            logging.getLogger(), state_file, jenkins)
        if prerequisites:
            for prerequisite in job.get_prerequisite_jobs():
                prerequisite.update_from_build_result('SUCCESS')
        return job

    def test_maybe_build(self):
        job = self.make_job()
        self.assertEqual(True, job.maybe_build([]))
        self.assertEqual(
            [((job.job_id, None), {})], job.jenkins.calls['build_job'])

    def test_maybe_build_conflicting(self):
        job = self.make_job()
        with patch.object(job, 'conflicting_jobs', return_value=True):
            self.assertEqual(False, job.maybe_build([]))
        self.assertNotIn('build_job', job.jenkins.calls)

    def test_maybe_build_prerequisites_conflict(self):
        job = self.make_job()
        active_jobs = [PUBLISH_REVISION]
        self.assertEqual(False, job.conflicting_jobs(active_jobs))
        self.assertEqual(False, job.maybe_build(active_jobs))

    def test_maybe_build_not_updates_state_file(self):
        job = self.make_job()
        old_status = dict(job.state_file._data)
        job.maybe_build([])
        self.assertEqual(old_status, job.state_file._data)

    def test_maybe_build_aborts_on_terminal_state_file_job(self):
        success_job = self.make_job()
        success_sfj = success_job.get_state_file_job()
        success_sfj.update_from_build_result('SUCCESS')
        self.assertEqual(False, success_job.maybe_build([]))
        failure_job = self.make_job()
        failure_sfj = failure_job.get_state_file_job()
        finally_fail_job(self, failure_sfj)
        self.assertEqual(False, failure_job.maybe_build([]))

    def test_maybe_build_needs_prerequisite_success(self):
        job = self.make_job(prerequisites=False)
        self.assertEqual(False, job.maybe_build([]))
        for prerequisite in job.get_prerequisite_jobs():
            prerequisite.update_from_build_result('SUCCESS')
        self.assertEqual(True, job.maybe_build([]))

    def test_maybe_build_calls_build(self):
        job = self.make_job()
        with patch.object(job, 'build') as build_mock:
            job.maybe_build([])
        build_mock.assert_called_with(None)


class TestBuildRevisionJob(TestCase):

    @contextmanager
    def make_job(self, revision=True):
        rev_to_build = (33, 'rev33') if revision else (None, None)
        jenkins = FakeJenkins({}, current_version=23)
        state_file = StateFile()
        state_file.start_revision('a', 1, 'b', '1.6', 2)
        logger = logging.getLogger('cidirector')
        job = BuildRevisionJob('foo', logger, state_file, jenkins)
        with patch.object(job, 'get_revision_to_build',
                          return_value=rev_to_build):
            with patch.object(job, 'build', return_value=43):
                yield job

    def test_maybe_build_builds(self):
        active_jobs = make_server_info('idle').get_building_jobs()
        with self.make_job() as job:
            self.assertEqual(True, job.maybe_build(active_jobs))
            job.build.assert_called_with(
                {'branch': 'foo', 'revision': 33})
        self.assertEqual(job.state_file.current_version_number, 2)

    def test_maybe_build_no_revision(self):
        active_jobs = make_server_info('idle').get_building_jobs()
        with self.make_job(revision=False) as job:
            self.assertEqual(False, job.maybe_build(active_jobs))
            self.assertEqual(0, job.build.call_count)
        self.assertEqual(job.state_file.current_version_number, 2)

    def test_maybe_build_testing(self):
        active_jobs = make_server_info('testing').get_building_jobs()
        with self.make_job() as job:
            self.assertEqual(False, job.maybe_build(active_jobs))
            self.assertEqual(0, job.build.call_count)
        self.assertEqual(job.state_file.current_version_number, 2)

    def test_maybe_build_preparing(self):
        active_jobs = make_server_info('preparing').get_building_jobs()
        with self.make_job() as job:
            self.assertEqual(False, job.maybe_build(active_jobs))
            self.assertEqual(0, job.build.call_count)
        self.assertEqual(job.state_file.current_version_number, 2)

    def test_get_bzr_revision_to_build_branch(self):
        with bzr_sandbox() as bzr_home:
            foo_tree, foo_path, foo_rev = make_bzr_tree(bzr_home, 'foo',
                                                        '1.23.5', 2)
            job = BuildRevisionJob(foo_path, logging.getLogger(),
                                   StateFile(), None)
            self.assertEqual((2, foo_rev),
                             job.get_bzr_revision_to_build())

    def test_get_bzr_revision_to_build_skips_seen_revisions(self):
        with bzr_sandbox() as bzr_home:
            bar_tree, bar_path, bar_rev = make_bzr_tree(bzr_home, 'bar',
                                                        '1.7', 2)
            job = BuildRevisionJob(
                bar_path, logging.getLogger(), StateFile(), None)
            self.assertEqual((2, bar_rev),
                             job.get_bzr_revision_to_build())
            job.state_file.start_revision('asdf', 7,
                                          bar_tree.last_revision(), '1.6', 8)
            self.assertEqual((None, None),
                             job.get_bzr_revision_to_build())

    def test_get_git_revision_to_build_branch(self):
        job = BuildRevisionJob('gitbranch:foo:github.com/bar/baz',
                               logging.getLogger(), StateFile(), None)
        with patch('cidirector.cidirector.GitRepo.list_remote_branch_tips',
                   return_value=[('foo', 'asdf')]):
            self.assertEqual(('asdf', 'asdf'),
                             job.get_git_revision_to_build())

    def test_get_git_revision_to_build_skips_seen_revisions(self):
        job = BuildRevisionJob('gitbranch:foo:github.com/bar/baz',
                               logging.getLogger(), StateFile(), None)
        with patch('cidirector.cidirector.GitRepo.list_remote_branch_tips',
                   return_value=[('foo', 'asdf')]):
            self.assertEqual(('asdf', 'asdf'),
                             job.get_git_revision_to_build())
            job.state_file.start_revision('asdf', 7, 'asdf', '1.6', 8)
            self.assertEqual((None, None), job.get_revision_to_build())

    def test_conflicting_jobs_includes_resourceful_jobs(self):
        jenkins = FakeJenkins({
            'foo-test': {
                'build_number': 5,
                'description': '[ci-director]\n requires: build-revision'},
        })
        job = BuildRevisionJob(None, None, None, jenkins)
        self.assertEqual(True, job.conflicting_jobs(['foo-test']))

    def test_adopt_build(self):
        job = BuildRevisionJob(None, logging.getLogger(), StateFile(),
                               FakeJenkins({}))
        build_info = make_build_info(57, branch='lp:foo', revision=23)
        job.jenkins.buildvars[WORKSPACE] = dedent("""\
            export BRANCH=lp:bar
            export REVNO=24
            export REVISION_ID='rev-24'
            export VERSION='1.17.2'
        """)
        job.adopt_build(build_info)
        del job.state_file._current_version['started']
        self.assertEqual(job.state_file._current_version, {
            'branch': 'lp:bar',
            'build': {
                'failures': 0, 'last_build': 57, 'status': 'pending',
                'failure-threshold': 1, 'vote': True, 'building': True},
            'release_number': '1.17.2',
            'revision_id': 'rev-24',
            'revno': 24,
            'tests': {}
        })

    def test_adopt_build_empty_buildvars(self):
        job = BuildRevisionJob(None, logging.getLogger(), StateFile(),
                               FakeJenkins({}))
        build_info = make_build_info(57, branch='lp:foo', revision=23)
        job.jenkins.buildvars[WORKSPACE] = None
        job.adopt_build(build_info)
        self.assertIs(job.state_file._current_version, None)

    def test_adopt_build_old_build(self):
        job = BuildRevisionJob(None, logging.getLogger(), StateFile(),
                               FakeJenkins({}))
        build_info = make_build_info(57, branch='lp:foo', revision=23)
        job.jenkins.buildvars[WORKSPACE] = dedent("""\
            export BRANCH=lp:bar
            export REVNO=24
            export REVISION_ID='rev-24'
            export VERSION='1.17.2'
        """)
        job.state_file.start_revision('lp:baz', 23, 'rev-23', '1.17.1', 58)
        job.adopt_build(build_info)
        del job.state_file._current_version['started']
        self.assertEqual(job.state_file._current_version, {
            'branch': 'lp:baz',
            'build': {
                'failures': 0, 'last_build': 58, 'status': 'pending',
                'failure-threshold': 1, 'vote': True, 'building': True},
            'release_number': '1.17.1',
            'revision_id': 'rev-23',
            'revno': 23,
            'tests': {}
        })

    def test_adopt_build_new_build(self):
        job = BuildRevisionJob(None, logging.getLogger(), StateFile(),
                               FakeJenkins({}))
        build_info = make_build_info(57, branch='lp:foo', revision=23)
        job.jenkins.buildvars[WORKSPACE] = dedent("""\
            export BRANCH=lp:bar
            export REVNO=24
            export REVISION_ID='rev-24'
            export VERSION='1.17.2'
        """)
        job.state_file.start_revision('lp:baz', 23, 'rev-23', '1.17.1', 56)
        job.adopt_build(build_info)
        del job.state_file._current_version['started']
        self.assertEqual(job.state_file._current_version, {
            'branch': 'lp:bar',
            'build': {
                'failures': 0, 'last_build': 57, 'status': 'pending',
                'failure-threshold': 1, 'vote': True, 'building': True},
            'release_number': '1.17.2',
            'revision_id': 'rev-24',
            'revno': 24,
            'tests': {}
        })

    def test_adopt_build_no_revno(self):
        job = BuildRevisionJob(None, logging.getLogger(), StateFile(),
                               FakeJenkins({}))
        build_info = make_build_info(57, branch='lp:foo', revision=23)
        job.jenkins.buildvars[WORKSPACE] = dedent("""\
            export BRANCH=lp:bar
            export REVISION_ID='rev-24'
            export VERSION='1.17.2'
        """)
        job.state_file.start_revision('lp:baz', 23, 'rev-23', '1.17.1', 56)
        job.adopt_build(build_info)
        del job.state_file._current_version['started']
        self.assertEqual(job.state_file._current_version, {
            'branch': 'lp:bar',
            'build': {
                'failures': 0, 'last_build': 57, 'status': 'pending',
                'failure-threshold': 1, 'vote': True, 'building': True},
            'release_number': '1.17.2',
            'revision_id': 'rev-24',
            'revno': None,
            'tests': {}
        })


class TestPublishRevisionResourcefulJob(TestCase):

    @staticmethod
    def get_test_job(prev_result=None, built=True):
        state_file = StateFile()
        state_file.start_revision('a', 1, 'b', '1.6', 5)
        if built:
            build_revision = state_file.get_build_revision_job()
            build_revision.update_from_build_result('SUCCESS')
        if prev_result is not None:
            pub_job = state_file.get_grouped_job(PUBLISH_REVISION)
            pub_job.start_build(7)
            pub_job.update_from_build_result(prev_result)
        jenkins = FakeJenkins({
            PUBLISH_REVISION: {
                'build_number': 2,
                'description': '[ci-director]\n requires: build-revision'}},
            current_version=5)
        return ResourcefulJob(
            PUBLISH_REVISION, logging.getLogger(), state_file, jenkins)

    def test_maybe_build(self):
        job = self.get_test_job()
        with patch.object(job, 'build') as build_mock:
            self.assertEqual(True, job.maybe_build([]))
        build_mock.assert_called_with({'revision_build': 5})
        self.assertIs(None,
                      job.state_file.publication_job()._data.get('last_build'))

    def test_maybe_build_published(self):
        job = self.get_test_job('SUCCESS')
        with patch.object(job, 'build') as build_mock:
            self.assertEqual(False, job.maybe_build([]))
        self.assertEqual(0, build_mock.call_count)
        self.assertEqual(7,
                         job.state_file.publication_job()._data['last_build'])

    def test_maybe_build_failed(self):
        job = self.get_test_job('FAILURE')
        with patch.object(job, 'build') as build_mock:
            self.assertEqual(False, job.maybe_build([]))
        self.assertEqual(0, build_mock.call_count)
        self.assertEqual(7,
                         job.state_file.publication_job()._data['last_build'])

    def test_maybe_build_aborted(self):
        job = self.get_test_job('ABORTED')
        with patch.object(job, 'build') as build_mock:
            self.assertEqual(True, job.maybe_build([]))
        build_mock.assert_called_with({'revision_build': 5})
        self.assertEqual(7,
                         job.state_file.publication_job()._data['last_build'])

    def test_maybe_build_not_built(self):
        job = self.get_test_job('NOT_BUILT')
        with patch.object(job, 'build', return_value=8) as build_mock:
            self.assertEqual(True, job.maybe_build([]))
        build_mock.assert_called_with({'revision_build': 5})
        self.assertEqual(7,
                         job.state_file.publication_job()._data['last_build'])

    def test_maybe_build_no_build_revision(self):
        job = self.get_test_job(built=False)
        with patch.object(job, 'build') as build_mock:
            self.assertEqual(False, job.maybe_build([]))
            self.assertEqual(build_mock.call_count, 0)
            build_revision = job.state_file.get_build_revision_job()
            build_revision.update_from_build_result('SUCCESS')
            self.assertEqual(True, job.maybe_build([]))
            build_mock.assert_called_with({'revision_build': 5})


class TestResourcefulJob(TestCase):

    def test_init(self):
        state_file = StateFile()
        job = ResourcefulJob(
            'foo', logging.getLogger(), state_file, FakeJenkins({}))
        self.assertEqual('foo', job.job_id)
        self.assertEqual(set(), job.requires)
        self.assertEqual(set(), job.conflicts)
        self.assertIs(None, job.group_name)
        self.assertIs(None, job.tags)
        self.assertEqual(1, job.failure_threshold)

    def test_init_with_configuration(self):
        job = ResourcefulJob(
            'foo', None, None, None,
            requires='foo', conflicts='bar baz', group_name='qux',
            tags='a=b c=d', failure_threshold=2)
        self.assertEqual(set(['foo']), job.requires)
        self.assertEqual(set(['bar', 'baz']), job.conflicts)
        self.assertIs('qux', job.group_name)
        self.assertIs('a=b c=d', job.tags)
        self.assertEqual(2, job.failure_threshold)

    def test_config_from_description(self):
        config = ResourcefulJob.config_from_description('id', None, None)
        self.assertIs(None, config)
        config = ResourcefulJob.config_from_description('id', '', None)
        self.assertIs(None, config)
        config = ResourcefulJob.config_from_description('id', 'bar', None)
        self.assertIs(None, config)
        # The [ci-director] directive must be the entire line.
        config = ResourcefulJob.config_from_description(
            'id', '#[ci-director]', None)
        self.assertIs(None, config)
        # The config is empty when there are no options found.
        config = ResourcefulJob.config_from_description(
            'id', 'foo\n[ci-director]\nskip', None)
        self.assertEquals({}, config)
        # The config only contains keys that match the supported options.
        # Just prefixing a '#' to an option causes a non-match.
        config = ResourcefulJob.config_from_description(
            'id', '[ci-director] \n requires: foo \n # conflicts: bar', None)
        self.assertEquals({'requires': 'foo'}, config)
        # The failure_threshold is set to 1 when the value cannot be parsed
        # to an int.
        config = ResourcefulJob.config_from_description(
            'id', '[ci-director] \n failure-threshold: 3 # not int',
            logging.getLogger())
        self.assertEquals({'failure_threshold': 1}, config)
        # Tags are supportes
        config = ResourcefulJob.config_from_description(
            'id', '[ci-director] \n tags: subject=foo,bar arch=baz',
            logging.getLogger())
        self.assertEquals({'tags': 'subject=foo,bar arch=baz'}, config)

    def test_get_job_from_jenkins(self):
        server_info = make_server_info(
            'idle', buildable=['foo'], building=['build-revision'],
            disabled=[])
        jenkins = FakeJenkins({
            'build-revision': {'build_number': 7, 'description': ''},
            'foo': {
                'build_number': 7,
                'description': '[ci-director]\n requires: build-revision'},
        })
        state_file = StateFile()
        state_file.start_revision('b', 1, 'id', '1.20,0', 7)
        logger = logging.getLogger()
        job_id = server_info.get_enabled_jobs()[0]
        job = ResourcefulJob.get_job_from_jenkins(
            job_id, jenkins, state_file, logger)
        self.assertIsInstance(job, ResourcefulJob)
        self.assertEqual('foo', job.job_id)
        with patch.object(JobInfo, 'from_jenkins',
                          side_effect=JenkinsException):
            job = ResourcefulJob.get_job_from_jenkins(
                job_id, jenkins, state_file, logger)
            self.assertIs(None, job)

    def test_get_available_jobs(self):
        server_info = make_server_info(
            'idle', buildable=['foo', 'bar', 'baz'],
            building=['build-revision'], disabled=[])
        jenkins = FakeJenkins({
            'build-revision': {'build_number': 7, 'description': ''},
            'foo': {'build_number': 7, 'description': ''},
            'bar': {'build_number': 7, 'description': '[ci-director]'},
            'baz': {
                'build_number': 7,
                'description': '[ci-director]\n requires: bar'},
        })
        state_file = StateFile()
        state_file.start_revision('b', 1, 'id', '1.20,0', 7)
        logger = logging.getLogger()
        available_jobs = ResourcefulJob.get_available_jobs(
            server_info, jenkins, state_file, logger)
        self.assertEqual(['baz'], [j.job_id for j in available_jobs])

    def test_get_build_parameters(self):
        state_file = StateFile()
        state_file.start_revision('b', 1, 'id', '1.20.0', 7)
        job = ResourcefulJob('foo', None, state_file, None)
        self.assertEqual({'revision_build': 7}, job.get_build_parameters())

    def test_conflicting_jobs(self):
        job = ResourcefulJob('foo', None, None, None, conflicts='baz')
        self.assertEqual(False, job.conflicting_jobs([]))
        self.assertEqual(
            True, job.conflicting_jobs([BuildRevisionJob.job_id]))
        self.assertEqual(True, job.conflicting_jobs(['baz']))

    def test_get_prerequisite_jobs(self):
        state_file = StateFile()
        state_file.start_revision('b', 1, 'id', '1.20,0', 7)
        job = ResourcefulJob(
            'foo', logging.getLogger(), state_file, None, requires='baz')
        # All ResourcefulJobs require the build job.
        # When a required job is not found, it is represented as None.
        self.assertEqual(
            [None, state_file.get_build_revision_job()],
            job.get_prerequisite_jobs())
        # When a required job is found, it is included.
        state_file.get_grouped_job('baz', 'fnord', 2)
        state_file.get_grouped_job('bar', 'fnord', 2)
        job_ids = [j.test_id for j in job.get_prerequisite_jobs()]
        self.assertEqual(['baz', 'build-revision'], job_ids)

    def test_get_state_file_job(self):
        state_file = StateFile()
        state_file.start_revision('b', 1, 'id', '1.20.0', 7)
        job = ResourcefulJob('foo', None, state_file, None, "bar=baz")
        self.assertEqual(
            state_file.get_grouped_job('foo'), job.get_state_file_job())

    def test_maybe_build(self):
        state_file = StateFile()
        state_file.start_revision('b', 1, 'id', '1.20,0', 7)
        jenkins = FakeJenkins({
            'foo': {'build_number': 7},
        })
        job = ResourcefulJob(
            'foo', logging.getLogger(), state_file, jenkins,
            requires='baz', conflicts='bar')
        req_job = state_file.get_grouped_job('baz', 'fnord', 2)
        con_job = state_file.get_grouped_job('bar', 'fnord', 2)
        # The required job has not succeeded.
        self.assertIs(False, job.maybe_build([]))
        # The required job has succeeded but because of a race, it is listed
        # in active jobs.
        state_file.get_build_revision_job().update_from_build_result('SUCCESS')
        req_job.update_from_build_result('SUCCESS')
        self.assertIs(False, job.maybe_build([req_job.test_id]))
        # The required job is now done, but the conflicting job is running.
        self.assertIs(False, job.maybe_build([con_job.test_id, 'snarf']))
        # Requirements are met and there are not active conflicting jobs.
        with patch.object(job, 'build') as mock:
            self.assertIs(True, job.maybe_build(['snarf']))
            self.assertEqual(1, mock.call_count)


class TestGetBuildvarsFile(TestCase):

    def assert_request_url(self, jo_mock, url):
        self.assertEqual(1, len(jo_mock.call_args[0]))
        self.assertEqual(jo_mock.call_args[0][0].get_full_url(), url)

    def test_get_buildvars_file_workspace(self):
        jenkins = Jenkins('http://foo')
        expected = object()
        with patch.object(jenkins, 'jenkins_open',
                          return_value=expected) as jo_mock:
            actual = get_buildvars_file(jenkins, {'building': True})
        self.assert_request_url(
            jo_mock,
            'http://foo/job/build-revision/ws/buildvars.bash')
        self.assertIs(actual, expected)

    def test_get_buildvars_file_archived(self):
        jenkins = Jenkins('http://foo')
        expected = object()
        with patch.object(jenkins, 'jenkins_open',
                          return_value=expected) as jo_mock:
            actual = get_buildvars_file(jenkins, {
                'building': False,
                'number': 1076,
            })
        self.assert_request_url(
            jo_mock,
            'http://foo/job/build-revision/1076/artifact/buildvars.bash')
        self.assertIs(actual, expected)


class TestParseBuildvarsFile(TestCase):

    def test_parse_buildvars_file(self):
        result = parse_buildvars_file(dedent("""\
            export BRANCH=lp:foo
            export REVNO=35
            export REVISION_ID='asd fz'
            export VERSION=1.16.27
        """))
        self.assertEqual(result, {
            'BRANCH': 'lp:foo',
            'REVNO': '35',
            'REVISION_ID': 'asd fz',
            'VERSION': '1.16.27',
        })


class TestMain(TestCase):

    def setUp(self):
        self.logger = logging.getLogger()
        self.orig_handlers = self.logger.handlers
        self.logger.handlers = []
        self.orig_level = self.logger.level

    def tearDown(self):
        self.logger.handlers = self.orig_handlers
        self.logger.level = self.orig_level

    def test_main_default(self):
        with patch('cidirector.cidirector.build_revision') as br_mock:
            with patch('cidirector.cidirector.load_plugins') as lp_mock:
                with patch('cidirector.cidirector.RotatingFileHandler'
                    ) as rfh_mock:
                    main(['b1', 'b2'])
                br_mock.assert_called_with(['b1', 'b2'])
                self.assertEqual(1, br_mock.call_count)
                self.assertEqual(1, lp_mock.call_count)
        root_logger = logging.getLogger()
        self.assertEqual(logging.WARNING, root_logger.level)
        rfh_mock.assert_called_once_with(
            'ci-director.log', backupCount=2, maxBytes=1024 * 1024)
        self.assertIs(sys.stderr, root_logger.handlers[1].stream)

    def test_main_logging_config(self):
        with patch('cidirector.cidirector.build_revision'):
            with patch('cidirector.cidirector.load_plugins'):
                with temp_dir() as log_dir:
                    log_path = '{}/ci-director.log'.format(log_dir)
                    main(['-v', '--log-path', log_path, '--log-count',
                          '5', 'b1'])
        root_logger = logging.getLogger()
        self.assertEqual(logging.INFO, root_logger.level)
        self.assertEqual(5, root_logger.handlers[0].backupCount)
        self.assertEqual(log_path, root_logger.handlers[0].baseFilename)
