__metaclass__ = type

from datetime import datetime
import logging
import os.path
from StringIO import StringIO
import subprocess
from unittest import TestCase

from mock import patch
import yaml

from cidirector.jobs import (
    JENKINS_BUILD_RESULTS,
    PUBLISH_REVISION,
)
from cidirector.storage import (
    BLESS,
    BUILDING,
    CURSE,
    FAILED,
    open_rw,
    PENDING,
    StateFile,
    StateFileJob,
    SUCCEEDED,
)
from cidirector.tests.test_cidirector import FakeJenkins
from cidirector.tests.utility import (
    finally_fail_job,
)
from cidirector.utility import (
    temp_dir,
)


ISO_8601_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'


class TestOpenRW(TestCase):

    def setUp(self):
        self.temp_dir_context = temp_dir()
        self.temp_dir = self.temp_dir_context.__enter__()
        self.filename = os.path.join(self.temp_dir, 'state-file')

    def tearDown(self):
        self.temp_dir_context.__exit__(None, None, None)

    def test_open_rw_creates(self):
        self.assertFalse(os.path.exists(self.filename))
        with open_rw(self.filename):
            pass
        self.assertTrue(os.path.exists(self.filename))

    def test_file_writable(self):
        with open_rw(self.filename) as test_file:
            test_file.write('foo')

    def test_file_readable(self):
        with open_rw(self.filename) as test_file:
            test_file.read()

    def test_no_truncate(self):
        with open(self.filename, 'w') as test_file:
            test_file.write('foo')
        with open_rw(self.filename) as test_file:
            self.assertEqual(test_file.read(), 'foo')


class TestStateFileJob(TestCase):

    def test_start_build(self):
        sfj = StateFileJob('foo', {})
        sfj.start_build(59)
        self.assertEqual(
            sfj._data, {
                'last_build': 59, 'status': PENDING, 'failures': 0,
                'failure-threshold': 1, 'vote': True, 'building': True})

    def test_start_build_keeps_failures(self):
        sfj = StateFileJob(
            'foo', {'last_build': 49, 'status': PENDING, 'failures': 1})
        sfj.start_build(59)
        self.assertEqual(
            sfj._data, {
                'last_build': 59, 'status': PENDING, 'failures': 1,
                'failure-threshold': 1, 'vote': True, 'building': True})

    def start_build_keeps_success(self):
        sfj = StateFileJob(
            'foo',
            {'last_build': 49, 'status': SUCCEEDED, 'failures': 1,
             'building': False})
        sfj.start_build(50)
        self.assertEqual(
            sfj._data, {
                'last_build': 50, 'status': SUCCEEDED, 'failures': 1,
                'failure-threshold': 1, 'vote': True, 'building': True})

    def test_update_from_build_result_failed(self, status='FAILURE'):
        sfj = StateFileJob('foo', {})
        sfj.start_build(59)
        sfj.update_from_build_result(status)
        self.assertEqual(
            sfj._data,
            {'last_build': 59, 'status': PENDING, 'failures': 1,
             'failure-threshold': 1, 'vote': True, 'building': False})

    def test_update_from_build_result_unstable(self):
        return self.test_update_from_build_result_failed('UNSTABLE')

    def test_update_from_build_result_until_test_failed(self,
                                                        status='FAILURE'):
        sfj = StateFileJob('foo', {})
        sfj.start_build(55)
        while sfj.get_status() != FAILED:
            sfj.update_from_build_result(status)
        status = sfj._data['status']
        self.assertEqual(status, PENDING)

    def test_update_from_build_result_until_test_unstable(self):
        return self.test_update_from_build_result_until_test_failed('UNSTABLE')

    def test_update_from_build_result_accepts_only_valid(self):
        sfj = StateFileJob('foo', {})
        for value in JENKINS_BUILD_RESULTS:
            sfj.start_build(3)
            sfj.update_from_build_result(value)
        with self.assertRaises(ValueError):
            sfj.update_from_build_result('INVALID')

    def test_update_from_build_result_aborted(self, status='ABORTED'):
        sfj = StateFileJob('foo', {})
        for value in JENKINS_BUILD_RESULTS:
            sfj.start_build(3)
            sfj.update_from_build_result(status)
        self.assertEqual(sfj._data,
                         {'status': PENDING, 'failures': 0, 'last_build': 3,
                          'failure-threshold': 1, 'vote': True,
                          'building': False})

    def test_update_from_build_result_not_built(self):
        self.test_update_from_build_result_aborted('NOT_BUILT')

    def test_get_status(self):
        sfj = StateFileJob('foo', {}, 5)
        self.assertIs(sfj.get_status(), PENDING)
        sfj.start_build(55)
        self.assertEqual(sfj.get_status(), BUILDING)
        sfj.update_from_build_result('FAILURE')
        self.assertEqual(sfj.get_status(), PENDING)
        finally_fail_job(self, sfj)
        self.assertEqual(sfj.get_status(), FAILED)

    def test_get_status_build_after_failure(self):
        sfj = StateFileJob('foo', {'failures': 0}, 5)
        finally_fail_job(self, sfj)
        sfj.start_build(6)
        self.assertEqual(sfj.get_status(), BUILDING)

    def test_get_status_extra_build_after_success(self):
        sfj = StateFileJob(
            'foo',
            {'failures': 0, 'last_build': 63, 'status': SUCCEEDED,
             'building': False},
            5)
        sfj.start_build(64)
        self.assertEqual(sfj.get_status(), BUILDING)
        sfj.update_from_build_result('FAILURE')
        # The result SUCCEDED is not overwritten by the new failure.
        self.assertEqual(sfj.get_status(), SUCCEEDED)

    def test_init_with_failure_threshold(self):
        data = {'status': PENDING, 'failures': 0, 'failure-threshold': 3}
        # The failure_threshold can come from data.
        data['failure-threshold'] = 3
        sfj = StateFileJob('foo', data, None)
        self.assertEqual(3, sfj.failure_threshold)
        self.assertEqual(3, sfj._data['failure-threshold'])
        self.assertEqual(0, sfj._data['failures'])
        # The failure_threshold in the data takes precedence over the arg
        # passed to init.
        sfj = StateFileJob('foo', data, 5)
        self.assertEqual(3, sfj.failure_threshold)
        self.assertEqual(3, sfj._data['failure-threshold'])
        self.assertEqual(0, sfj._data['failures'])
        # When neither the data doesn't care, the value comes from the arg
        del data['failure-threshold']
        sfj = StateFileJob('foo', data, 5)
        self.assertEqual(5, sfj.failure_threshold)
        self.assertEqual(5, sfj._data['failure-threshold'])
        self.assertEqual(0, sfj._data['failures'])
        # When neither the data and the failure_threshold arg doesn't care,
        # the value is set to 1
        del data['failure-threshold']
        sfj = StateFileJob('foo', data, None)
        self.assertEqual(1, sfj.failure_threshold)
        self.assertEqual(1, sfj._data['failure-threshold'])

    def test_update_from_jenkins(self):
        jenkins = FakeJenkins({
            'foo': {
                'build_number': 2, 'building': False, 'result': 'SUCCESS'},
        })
        sfj = sfj = StateFileJob('foo', {'last_build': 2}, None)
        sfj.update_from_jenkins(jenkins, 12)
        self.assertEqual('succeeded', sfj._data['status'])

    def test_update_from_jenkins_adapts_specialised_jobs(self):
        jenkins = FakeJenkins({
            PUBLISH_REVISION: {
                'build_number': 2, 'building': False, 'result': 'SUCCESS'},
        })
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20.1', 2)
        for method in [sf.publication_job]:
            sfj = sf.publication_job()
            sfj._data['last_build'] = 2
            sfj.update_from_jenkins(jenkins, 2)
            self.assertEqual('succeeded', sfj._data['status'])

    def test_update_from_jenkins_triggers_archiving_artifacts(self):
        build_number = 1234
        foo_job = 'foo'
        foo_data = {
            'status': PENDING, 'failures': 0, 'failure-threshold': 3,
            'build_number': build_number, 'last_build': build_number,
            'building': True,
        }
        jobs = {foo_job: foo_data}
        current_version = 4567
        jenkins = FakeJenkins(jobs, current_version)
        sfj = StateFileJob(foo_job, foo_data, None)
        with patch.object(sfj, 'archive_results') as mocked_archive_results:
            self.assertEqual(BUILDING, sfj.get_status())
            sfj.update_from_jenkins(jenkins, current_version)
            self.assertEqual(BUILDING, sfj.get_status())
            self.assertEqual(0, len(mocked_archive_results.mock_calls))

            artifacts = [
                {
                    'fileName': 'some-product.tar.gz',
                    'relativePath': 'build/some-product.tar.gz',
                },
            ]
            jenkins.set_build_result(
                foo_job, build_number, 'SUCCESS', artifacts)
            sfj.update_from_jenkins(jenkins, current_version)
            self.assertEqual(SUCCEEDED, sfj.get_status())
            mocked_archive_results.assert_called_once_with(
                artifacts, current_version, 'fakettp://fake.fake/job/foo/1234',
                build_number)

    def test_update_from_jenkins_missing_build(self):
        build_number = 1234
        foo_job = 'foo'
        foo_data = {
            'status': PENDING, 'failures': 0, 'failure-threshold': 3,
            'build_number': build_number, 'last_build': build_number - 1,
            'building': True,
        }
        jobs = {foo_job: foo_data}
        current_version = 4567
        jenkins = FakeJenkins(jobs, current_version)
        logger = logging.getLogger()
        handler = logging.handlers.MemoryHandler(10)
        logger.addHandler(handler)
        sfj = StateFileJob(foo_job, foo_data, None, logger=logger)
        sfj.update_from_jenkins(jenkins, current_version)
        self.assertEqual(PENDING, sfj._data['status'])
        self.assertIs(None, sfj._data['last_build'])
        self.assertFalse(sfj._data['building'])
        self.assertEqual(
            'Build 1233 of job foo is missing', handler.buffer[0].msg)

    def test_archive_results(self):
        sfj = StateFileJob('foo', {})
        artifacts = [
            {
                'fileName': 'build.log',
                'relativePath': 'logs/build.log',
            },
            {
                'fileName': 'bar',
                'relativePath': 'bar',
            },
            {
                'fileName': 'empty',
                'relativePath': 'empty',
            },
        ]
        with patch.object(subprocess, 'check_call') as mocked_check_call:
            with patch.object(
                    subprocess, 'check_output') as mocked_check_output:
                sfj.archive_results(
                    artifacts, 1234, 'http://juju-ci.apour.ws:8080/job/foo/42',
                    4567)
        call_params = mocked_check_call.call_args_list
        # Retrieve the changing temp dir name from deeply nested sequences:
        # - call_params has one entry for each call;
        # - data for each call cosists of [args, kwargs]
        # - we want args[0], which itself is a list of command line
        #   args.
        temp_file = call_params[0][0][0][3]
        expected = ([
            'wget', '-q', '-O', temp_file,
            'http://juju-ci.apour.ws:8080/job/foo/42/artifact/logs/build.log',
        ], )
        self.assertEqual(expected, call_params[0][0])
        temp_file = call_params[1][0][0][3]
        expected = ([
            'wget', '-q', '-O', temp_file,
            'http://juju-ci.apour.ws:8080/job/foo/42/artifact/bar',
        ], )
        self.assertEqual(expected, call_params[1][0])
        local_dir = mocked_check_output.call_args_list[0][0][0][5]
        self.assertEqual('/', local_dir[-1])
        s3_config = os.path.join(
            os.environ['HOME'], 'cloud-city/juju-qa.s3cfg')
        expected = ([
            's3cmd', '-c', s3_config, '--no-progress',
            'sync', local_dir,
            's3://juju-qa-data/juju-ci/products/version-1234/foo/build-4567/',
        ])
        mocked_check_output.assert_called_once_with(expected)
        self.assertEqual(2, len(call_params))


class TestStateFile(TestCase):

    def test_file_name(self):
        with patch.dict(os.environ, {'HOME': '/home/jrandomx'}):
            self.assertEqual('/home/jrandomx/.config/ci-director-state',
                             StateFile.file_name())

    def test_acquire_no_file(self):
        with temp_dir() as directory:
            filename = os.path.join(directory, 'state-file')
            with patch.object(StateFile, 'file_name', return_value=filename):
                with StateFile.acquire():
                    self.assertTrue(os.path.exists(filename))

    def test_acquire_retains_contents(self):
        with temp_dir() as directory:
            filename = os.path.join(directory, 'state-file')
            with open(filename, 'w') as f:
                f.write('foo')
            with patch.object(StateFile, 'file_name', return_value=filename):
                with StateFile.acquire() as (sf, f):
                    f.seek(0)
                    self.assertEqual(f.read(), 'foo')

    def test_from_file(self):
        data = ['foo', 'bar', 'baz']
        fileobj = StringIO()
        yaml.safe_dump(data, fileobj)
        fileobj.seek(0)
        sf = StateFile.from_file(fileobj)
        self.assertEqual(data, sf._data)
        self.assertIs(fileobj, sf._fileobj)
        sf2 = StateFile.from_file(fileobj)
        self.assertEqual(data, sf2._data)

    def test_write_seeks_truncates(self):
        fileobj = StringIO()
        yaml.dump('foo', fileobj)
        foo_yaml = fileobj.getvalue()
        fileobj.write('trailing')
        sf = StateFile('foo', fileobj)
        sf.write()
        self.assertEqual(foo_yaml, fileobj.getvalue())

    def test_start_revision(self):
        sf = StateFile()
        self.assertEqual(sf._data, {'versions': {}, 'test_builds': {},
                                    'versions_by_revision': {}})
        before = datetime.utcnow()
        sf.start_revision('lp:foo', 25, 'revid-blah', '14.6.3', 57)
        after = datetime.utcnow()
        started = datetime.strptime(
            sf._data['versions'][57].pop('started'), ISO_8601_FORMAT)
        self.assertTrue(before <= started <= after)
        self.assertEqual(sf._data, {
            'current_version': 57,
            'versions': {
                57: {
                    'branch': 'lp:foo',
                    'revno': 25,
                    'revision_id': 'revid-blah',
                    'build': {
                        'failures': 0,
                        'last_build': 57,
                        'status': PENDING,
                        'failure-threshold': 1,
                        'vote': True,
                        'building': True,
                    },
                    'tests': {},
                    'release_number': '14.6.3',
                }
            },
            'test_builds': {},
            'versions_by_revision': {'revid-blah': [57]},
        })

    def test_finish_revision(self):
        sf = StateFile()
        sf.start_revision('a', 5, 'b', '0.21', 'c')
        self.assertIs(None, sf.get_current_version().get('finished'))
        sf.finish_revision(datetime(1, 2, 3, 4, 5, 6, 7))
        self.assertEqual(sf.get_current_version()['finished'],
                         '0001-02-03T04:05:06.000007Z')

    def test_get_current_version_summary(self):
        sf = StateFile()
        sf.start_revision('lp:foo', 25, 'revid-blah', '1.6', 57)
        self.assertEqual(sf.get_current_version_summary(), {
            'branch': 'lp:foo',
            'revno': 25,
            'short_revision': 'r25',
            'revision_id': 'revid-blah',
            'build_number': 57,
            'release_number': '1.6',
        })

    def test_get_current_version_summary_no_revno(self):
        sf = StateFile()
        sf.start_revision('lp:foo', None, '1234567890', '1.6', 57)
        self.assertEqual(sf.get_current_version_summary(), {
            'branch': 'lp:foo',
            'revno': None,
            'short_revision': '12345678',
            'revision_id': '1234567890',
            'build_number': 57,
            'release_number': '1.6',
        })

    def test_revision_prepared(self):
        sf = StateFile()
        self.assertIs(False, sf.revision_prepared('ab'))
        self.assertIs(False, sf.revision_prepared('cd'))
        sf.start_revision('a', 5, 'ab', '1.6', '6')
        self.assertIs(True, sf.revision_prepared('ab'))
        self.assertIs(False, sf.revision_prepared('cd'))

    def test_roundtrip(self):
        sf = StateFile(fileobj=StringIO())
        sf.start_revision('a', 5, 'ab', '1.6', '6')
        sf.write()
        sf._fileobj.seek(0)
        sf2 = StateFile.from_file(sf._fileobj)
        self.assertEqual(sf._data, sf2._data)

    def test_save_result_bless(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        sf.save_result(BLESS, 90)
        self.assertEqual(
            sf._current_version['result'], {'status': BLESS, 'build': 90})

    def test_save_result_curse(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        sf.save_result(CURSE, 90)
        self.assertEqual(
            sf._current_version['result'], {'status': CURSE, 'build': 90})

    def test_save_result_pending(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        with self.assertRaisesRegexp(ValueError, 'Illegal value: pending'):
            sf.save_result(PENDING, 90)

    def test_save_result_other(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        with self.assertRaisesRegexp(ValueError, 'Illegal value: other'):
            sf.save_result('other', 90)

    def test_get_result(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        self.assertEqual((None, None), sf.get_result())
        sf.save_result(CURSE, 90)
        self.assertEqual((CURSE, 90), sf.get_result())
        sf.save_result(BLESS, 123)
        self.assertEqual((BLESS, 123), sf.get_result())

    def test_publication_job(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        self.assertNotIn(PUBLISH_REVISION, sf._data['versions'][2])
        pubjob = sf.publication_job()
        self.assertIn(
            PUBLISH_REVISION,
            sf._data['versions'][2]['release-tests'])
        self.assertIs(PENDING, pubjob.get_status())

    def test_publication_job_is_resourceful(self):
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20.1', 2)
        self.assertNotIn(PUBLISH_REVISION, sf._data['versions'][2])
        job = sf.get_grouped_job(
            PUBLISH_REVISION, 'release-tests', {'marker': True})
        pub_job = sf.publication_job()
        self.assertEqual(job, pub_job)

    def test_publication_job_no_version(self):
        sf = StateFile()
        self.assertIs(None, sf.publication_job())

    def test_get_build_revision_job(self):
        sf = StateFile()
        self.assertIs(None, sf.get_build_revision_job())
        sf.start_revision('a', 1, 'b', '1.6', 2)
        self.assertEqual(
            sf.get_build_revision_job(),
            StateFileJob('build-revision', {
                'status': PENDING,
                'failure-threshold': 1,
                'failures': 0,
                'last_build': 2,
                'building': True,
            }))

    def test_get_grouped_job(self):
        # Calling get_grouped_job() without a group creates a job in
        # the anonymous top-level group with a failure_threshold of 1.
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20', 7)
        grouped_job = sf.get_grouped_job('foo')
        self.assertIn(
            ('foo', {'status': PENDING, 'tags': None, 'failures': 0,
                     'vote': True, 'failure-threshold': 1,
                     'last_build': None, 'building': False}),
            sf._data['versions'][7].items())
        self.assertEqual(grouped_job.failure_threshold, 1)
        finally_fail_job(self, grouped_job)
        self.assertIn(
            ('foo', {'status': PENDING, 'tags': None, 'failures': 1,
                     'vote': True, 'failure-threshold': 1,
                     'last_build': None, 'building': False}),
            sf._data['versions'][7].items())

    def test_get_grouped_job_with_group_with_failure_threshold(self):
        # Calling get_grouped_job() with a group creates both the group
        # if needed and the job. The failure_threshold is set.
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20', 7)
        grouped_job = sf.get_grouped_job('foo', 'baz', 2)
        self.assertIn(
            ('foo', {'status': PENDING, 'tags': None, 'failures': 0,
                     'vote': True, 'failure-threshold': 2,
                     'last_build': None, 'building': False}),
            sf._data['versions'][7]['baz'].items())
        self.assertEqual(grouped_job.failure_threshold, 2)
        finally_fail_job(self, grouped_job)
        self.assertIn(
            ('foo', {'status': PENDING, 'tags': None, 'failures': 2,
                     'vote': True, 'failure-threshold': 2,
                     'last_build': None, 'building': False}),
            sf._data['versions'][7]['baz'].items())

    def test_get_grouped_job_with_tags(self):
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20', 7)
        sf.get_grouped_job('foo', 'baz', 2, 'qux')
        self.assertIn(
            ('foo', {'status': PENDING, 'tags': 'qux', 'failures': 0,
                     'vote': True, 'failure-threshold': 2,
                     'last_build': None, 'building': False}),
            sf._data['versions'][7]['baz'].items())

    def test_iter_grouped_jobs(self):
        sf = StateFile()
        sf.start_revision('b', 1, 'id', '1.20,0', 7)
        self.assertItemsEqual([], sf.iter_grouped_jobs())
        sf.get_grouped_job('foo', 'baz', 2)
        sf.get_grouped_job('bar', None, 2)
        sf.save_result('curse', 7)
        data = {
            'status': PENDING, 'tags': None, 'failures': 0, 'vote': True,
            'failure-threshold': 2}
        self.assertItemsEqual([
            StateFileJob('foo', data, failure_threshold=None),
            StateFileJob('bar', data, failure_threshold=None),
        ], sf.iter_grouped_jobs())

    def test_get_set_complete_failure_info(self):
        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        info = sf.get_complete_failure_info(2)
        self.assertIs(None, info)
        sf.set_complete_failure_info(2, {'foo-test': 1, 'bar-test': 2})
        info = sf.get_complete_failure_info(2)
        self.assertEqual({'foo-test': 1, 'bar-test': 2}, info)

    def test_build_info_to_s3(self):
        s3cmd_call_data = []

        def catch_temp_file_contents(call_params):
            file_path = call_params[5]
            s3cmd_call_data.append(file_path)
            s3cmd_call_data.append(yaml.load(open(file_path)))

        sf = StateFile()
        sf.start_revision('a', 1, 'b', '1.6', 2)
        with patch.object(
                subprocess, 'check_output',
                side_effect=catch_temp_file_contents) as check_output_mock:
            sf.build_info_to_s3()
        s3_config = os.path.join(
            os.environ['HOME'], 'cloud-city/juju-qa.s3cfg')
        check_output_mock.assert_called_once_with([
            's3cmd', '-c', s3_config, '--no-progress',
            'put', s3cmd_call_data[0],
            's3://juju-qa-data/juju-ci/products/version-2/result.json'
        ])
        expected_build_result = sf.get_current_version()
        self.assertEqual(expected_build_result, s3cmd_call_data[1])
